"""
Requirement: Open Webull Desktop terminal
"""


import pyautogui
import pytesseract
import time
import os
import re
import pandas as pd
import numpy as np
from tqdm import tqdm
import cv2
import multitasking as _multitasking

from IPython.display import clear_output

# # cropped_img = img[0:60, 0:900]
# _ohlcv_yyxx = [60, 120, 0, 900]
# # cropped_img = img[-100:, -510:-360]
# _time_yyxx = [-50, None, -500, -310]

_ohlcv_yyxx = [85, 120, 17, 900]
_time_yyxx = [-112, -62, -510, -200] #[-112, -62, -480, -260]


def mk_screenshot():
    # make screen shot of ohlcv
    # have adjusted area
    # save root was set in Mac shift+cmd+5 options
    pyautogui.hotkey('shift', 'command', '5')
    pyautogui.hotkey('enter')

def drag_back_one_tick(_base_point=(1350, 220), _tick_gap=8):
    pyautogui.click(_base_point)
    pyautogui.mouseDown(_base_point, button='left')
    pyautogui.dragTo(_base_point[0]+_tick_gap, _base_point[1], button='left')

def re_identify_sid(fname):
    sid = "".join(
        re.findall("[0-9]", 
        re.findall("[0-9]{4}-[0-9]{2}-[0-9]{2} at [0-9]{2}.[0-9]{2}.[0-9]{2}", fname)[0]))
    return sid

# def re_identify_ohlcv_elements(ohlcvstr):
#     elements = re.findall("[.0-9%-]+", ohlcvstr.replace(",", ""))
#     elements = [e for e in elements if len(e)>1]
#     return elements
def re_identify_ohlcv_elements(ohlcvstr):
    elements = re.findall("([-]{,1}[0-9]+[.][0-9]{2}[%]{,1})", ohlcvstr.replace(",", ""))
    elements = [e for e in elements if len(e)>1]
    return elements

# def re_identify_timestr(time_str):
#     time_str = re.findall("[0-9]{2}/[0-9]{2} [0-9]{2}:[0-9]{2}", time_str)[0]
#     return time_str
def re_identify_timestr(time_str):
    time_str = re.findall("([0-9]{2}/[0-9]{2})[ ]{,1}([0-9]{2}:[0-9]{2})", time_str)[0]
    time_str = " ".join(time_str)
    time_str = re.findall("[0-9]{2}/[0-9]{2} [0-9]{2}:[0-9]{2}", time_str)[0]
    return time_str

def dispatch_ohlcv_time(
        froot, 
        distfname='str.csv',
        recordstr=True,
        multi=True,
        _flder_suffix_dict = {
        'origin': 'origin/',
        'ohlcv': 'ohlcv/',
        'timelog': 'timelog/',
        'processed': 'bin/'
    }):
    fs = [f for f in os.listdir(froot + _flder_suffix_dict['origin']) if ".png" in f]
    if multi:
        for fname in tqdm(fs):
            try:
                dispatch_ohlcv_time_multi(fname, froot, distfname, recordstr, _flder_suffix_dict)
            except:
                time.sleep(10)
                dispatch_ohlcv_time_multi(fname, froot, distfname, recordstr, _flder_suffix_dict)

    else:
        for fname in tqdm(fs):
            dispatch_ohlcv_time_solo(fname, froot, distfname, recordstr, _flder_suffix_dict)
    return 

def dispatch_ohlcv_time_solo(
        fname,
        froot, 
        distfname='str.csv',
        recordstr=True,
        _flder_suffix_dict = {
        'origin': 'origin/',
        'ohlcv': 'ohlcv/',
        'timelog': 'timelog/',
        'processed': 'bin/'
    }, _ohlcv_yyxx=_ohlcv_yyxx, _time_yyxx=_time_yyxx):
    imgpath = froot + _flder_suffix_dict['origin'] + fname
    img = cv2.imread(imgpath)
    # ohlcv
    ohlcv_root = froot + _flder_suffix_dict['ohlcv']
    # cropped_img = img[0:60, 0:900]
    cropped_img = img[_ohlcv_yyxx[0]:_ohlcv_yyxx[1], _ohlcv_yyxx[2]:_ohlcv_yyxx[3]]
    cv2.imwrite(ohlcv_root+fname, cropped_img)
    # time
    time_root = froot + _flder_suffix_dict['timelog']
    # cropped_img = img[-100:, -510:-360]
    cropped_img = img[_time_yyxx[0]:_time_yyxx[1], _time_yyxx[2]:_time_yyxx[3]]
    cv2.imwrite(time_root+fname, cropped_img)

    if recordstr:
        try:
            record_str_from_ohlcv_time(froot, fname, distfname=distfname, _flder_suffix_dict=_flder_suffix_dict)
        except Exception as e:
            print(imgpath, e)
            
    return 

@_multitasking.task
def dispatch_ohlcv_time_multi(fname, froot, distfname, recordstr, _flder_suffix_dict):
    dispatch_ohlcv_time_solo(fname, froot, distfname, recordstr, _flder_suffix_dict)

def record_str_from_ohlcv_time(froot, fname, distfname, _flder_suffix_dict):
    time_root = froot + _flder_suffix_dict['timelog']
    ohlcv_root = froot + _flder_suffix_dict['ohlcv']

    time_str = screenshot_to_str(time_root+fname).replace("\n", "")
    time_str = re_identify_timestr(time_str)
    _date, _time = time_str.split(" ")

    move_from_a_to_b(froot + _flder_suffix_dict['origin'], froot + _flder_suffix_dict['processed'], fname)
    sid = re_identify_sid(fname)
    screenshot_to_strfile(ohlcv_root+fname, sid=sid, distfroot=froot, distfname=distfname, additional_cols=[_date, _time])
    return 

def record_str_from_full(froot, fname, distfname, _flder_suffix_dict,):
    full_root = froot + _flder_suffix_dict['processed']

    sid = re_identify_sid(fname)
    fullscreenshot_to_strfile(full_root+fname, sid=sid, distfroot=froot, distfname=distfname, )
    # move_from_a_to_b(froot + _flder_suffix_dict['origin'], froot + _flder_suffix_dict['processed'], fname)
    return 

# def dispatch_ohlcv_time_and_recordstrfile(froot, distfname="str.csv"):
#     fname = None
#     while fname is None:
#         try:
#             fname = [f for f in os.listdir(froot) if ".png" in f][0]
#         except:
#             time.sleep(1)
#     imgpath = froot + fname
#     img = cv2.imread(imgpath)
#     # ohlcv
#     ohlcv_root = froot + "ohlcv/"
#     cropped_img = img[0:60, 0:900]
#     cv2.imwrite(ohlcv_root+fname, cropped_img)
#     # ohlcv_str = screenshot_to_str(ohlcv_root+fname)
#     # time
#     time_root = froot + "timelog/"
#     cropped_img = img[-100:, -510:-360]
#     cv2.imwrite(time_root+fname, cropped_img)
#     time_str = screenshot_to_str(time_root+fname).replace("\n", "")
#     time_str = re_identify_timestr(time_str)
#     _date, _time = time_str.split(" ")

#     move_from_a_to_b(froot, froot + "origin/", fname)
#     sid = re_identify_sid(fname)
#     screenshot_to_strfile(ohlcv_root+fname, sid=sid, distfroot=froot, distfname=distfname, additional_cols=[_date, _time])
#     return 

def screenshot_to_str(img_path, ):
    return pytesseract.image_to_string(img_path)

def screenshot_to_strfile(img_path=None, sid='',
        distfroot="/Users/my/Documents/webull_tick_imags/20250101TSLA/", 
        distfname="str.csv", additional_cols=[]):
    xstr = screenshot_to_str(img_path)
    elements = re_identify_ohlcv_elements(xstr)
    elements.extend(additional_cols)
    elementstr = sid + "," + ",".join(elements) + '\n'
    with open(distfroot+distfname, 'a') as distf:
        distf.write(elementstr)

def re_identify_ohlcv_elements_full(ohlcvstr):
    elements = re.findall("[0OHLCV] ([0-9]{3}[.][0-9]{2})|([-+][0-9]+[.][0-9]{2}[%]{,1})|Vol ([0-9,]+)", ohlcvstr.replace(",", ""))
    from functools import reduce
    _e = reduce(lambda x,y: x+y, elements)
    elements = [e.replace("+", "") for e in _e if len(e)>1]
    timestr = re_identify_timestr(ohlcvstr)
    elements.extend(timestr.split(" "))
    return elements

def fullscreenshot_to_strfile(img_path=None, sid='',
        distfroot="/Users/my/Documents/webull_tick_imags/20250101TSLA/", 
        distfname="str.csv", ):
    xstr = screenshot_to_str(img_path)
    elements = re_identify_ohlcv_elements_full(xstr)
    elementstr = sid + "," + ",".join(elements) + '\n'
    with open(distfroot+distfname, 'a') as distf:
        distf.write(elementstr)
    
def imgfolder_to_strfile(
        img_root="/Users/my/Documents/webull_tick_imags/20250101TSLA/ohlcv/", 
        distfroot="/Users/my/Documents/webull_tick_imags/20250101TSLA/", distfname="str.csv", dispose=False):
    fs = [f for f in os.listdir(img_root) if ".png" in f]
    for f in tqdm(fs):
        sid = re_identify_sid(f)
        screenshot_to_strfile(img_root+f, sid, distfroot, distfname)
        if dispose:
            move_from_a_to_b(img_root, distfroot+"bin/", f)
            
def move_from_a_to_b(aroot, broot, ffullname):
    import shutil
    shutil.move(aroot+ffullname, broot+ffullname)


class TickCollector:
    def __init__(self, 
                 distfroot=None, 
                 distfname=None,
                 targetticknum=1000,
                 _base_point=(1350, 220), 
                 _tick_gap=8,
                 _flder_suffix_dict={
                    'origin': 'origin/',
                    'ohlcv': 'ohlcv/',
                    'timelog': 'timelog/',
                    'processed': 'bin/'
                }
            ):
        self.distfroot = distfroot
        self.distfname = distfname
        self.targetticknum = targetticknum
        self._flder_suffix_dict = _flder_suffix_dict
        
        self._base_point = _base_point
        self._tick_gap = _tick_gap

        self.doneticknum = 0

    def backtracker(self, flydispose=False, bucknum=50):
        self.img_root = self.distfroot + 'ohlcv/'
        self.flydispose = flydispose
        self.bucknum = bucknum
        self._done_bucks = 0
        self._total_bucks = int(np.ceil(self.targetticknum / self.bucknum))
        while self.doneticknum<self.targetticknum:
            self._backtracker_eachbuck()
            self._done_bucks += 1

    def _backtracker_eachbuck(self):
        # collect ticks
        print("Collecting ticks...")
        self.i = 0
        for i in tqdm(range(self.bucknum)):
            if self.doneticknum<self.targetticknum:
                self.update_state()
                self._screenshot()
                self.doneticknum+=1
                self.i += 1
                time.sleep(2)
        # imgs to strfile
        print("Transforming images...")
        imgfolder_to_strfile(img_root=self.img_root, 
            distfroot=self.distfroot, distfname=self.distfname, dispose=self.flydispose)
        
    # def flytracker(self):
    #     while self.doneticknum<self.targetticknum:
    #         self.update_state(buck=False)
    #         self._screenshot()
    #         dispatch_ohlcv_time_and_recordstrfile(self.distfroot, distfname=self.distfname)
    #         self.doneticknum += 1
    #         time.sleep(2)

    def pure_dispatch_ohlcv_time(self, multi=True):
        dispatch_ohlcv_time(
            froot=self.distfroot,
            distfname=self.distfname,
            recordstr=True,
            multi=multi,
            _flder_suffix_dict=self._flder_suffix_dict,
        )
        restnum = 100
        while restnum!=0:
            restnum = len([f for f in os.listdir(self.distfroot + self._flder_suffix_dict['origin']) if ".png" in f])
            clear_output(wait=True)
            print("{} image/s remain to be processed...".format(restnum))
        print('Done')
        return
    
    def pure_screenshot(self, sleep=1):
        while self.doneticknum<self.targetticknum:
            self.update_state(buck=False)
            self._screenshot()
            self.doneticknum += 1
            time.sleep(sleep)

    def _screenshot(self, ):
        drag_back_one_tick(_base_point=self._base_point, _tick_gap=self._tick_gap)
        mk_screenshot()
        
    def update_state(self, buck=False):
        clear_output(wait=True)
        if buck:
            print("[", "*"*self._done_bucks, "-"*(self._total_bucks-self._done_bucks), "]", "{:.2f}% Done".format(self._done_bucks/self._total_bucks), "{} ticks Done".format(self.doneticknum))
            print("|", "*"*self.i, "-"*(self.bucknum-self.i), "|")
        else:
            print("{} ticks Done".format(self.doneticknum))
            print("|", "*"*self.doneticknum, "-"*(self.targetticknum-self.doneticknum), "|")
    




from ibapi.client import *
from ibapi.wrapper import *
import datetime
import time
import threading

port = 7496


class DataApp(EClient, EWrapper):
    def __init__(self):
        EClient.__init__(self, self)
        self.orderId=0
        self.datapool = {}

    def nextValidId(self, orderId: OrderId):
        self.orderId = orderId
    
    def nextId(self):
        self.orderId += 1
        return self.orderId
    
    def error(self, reqId, errorCode, errorString, advancedOrderReject=""):
        print(f"reqId: {reqId}, errorCode: {errorCode}, errorString: {errorString}, orderReject: {advancedOrderReject}")
    
    def historicalData(self, reqId, bar):
        print(reqId, bar)
        if reqId not in self.datapool.keys():
            self.datapool[reqId] = []
        self.datapool[reqId].append(bar)
    
    def historicalDataEnd(self, reqId, start, end):
        print(f"Historical Data Ended for {reqId}. Started at {start}, ending at {end}")
        self.cancelHistoricalData(reqId)

# app = DataApp()
# app.connect("127.0.0.1", port, 0)
# threading.Thread(target=app.run).start()
# time.sleep(1)


def init_qqq():
    contractbase = Contract()
    contractbase.symbol = "QQQ"
    contractbase.secType = "STK"
    contractbase.exchange = "SMART"
    contractbase.currency = "USD"
    return contractbase

# contractx = init_qqq()


def init_qqq_options(ltd='20250527', rt='P', stk=510):
    contract1 = Contract()
    contract1.symbol = "QQQ"
    contract1.secType = "OPT"
    contract1.exchange = "SMART"
    contract1.currency = "USD"
    contract1.lastTradeDateOrContractMonth = ltd
    contract1.right = rt
    contract1.strike = stk
    return contract1

def init_tsla_options(ltd='20250527', rt='P', stk=510):
    contract1 = Contract()
    contract1.symbol = "TSLA"
    contract1.secType = "OPT"
    contract1.exchange = "SMART"
    contract1.currency = "USD"
    contract1.lastTradeDateOrContractMonth = ltd
    contract1.right = rt
    contract1.strike = stk
    return contract1


# import yfinance as yf

# qqqpx = yf.download('QQQ')
# qqqpxpd = qqqpx.xs('QQQ', axis=1, level=1)



def get_qqq_bardatapd(rawdatalist, _reqid, suffix=None, savepath=None, paramslog={}):
    import re
    import pandas as pd
    import os
    barsdata = re.findall(
    "([0-9]{2}[:][0-9]{2}[:][0-9]{2}) US/Eastern, Open: ([.0-9]+), High: ([.0-9]+), Low: ([.0-9]+), Close: ([.0-9]+), Volume: ([.0-9]+)", 
    " ".join([str(x) for x in rawdatalist]))
    _rt, _ltd, _bst, _ltddate, _stk = paramslog[_reqid]
    frq = _bst[1].replace(" ", "")
    fpath = f'{savepath}/{_ltddate}/'
    os.makedirs(fpath, exist_ok=True)
    if suffix is None:
        pd.DataFrame(barsdata).to_csv(f'{fpath}/0DTE{_rt}{_stk}_{frq}.csv')
    else:
        pd.DataFrame(barsdata).to_csv(f'{fpath}/0DTE{_rt}{_stk}_{frq}.csv')



# rtlist = ['C', 'P']
# ltdlist = [f"{xdt} 16:00:00 US/Eastern" for xdt in [
#     (datetime.datetime.today() - datetime.timedelta(x)).strftime("%Y%m%d") for x in range(1, 2)
# ]]
# barsetting = [("1 D", "1 min", "TRADES")]
# # barsetting = [("1 D", "5 mins", "TRADES")]

# from itertools import product

# _reqlist = list(product(rtlist, ltdlist, barsetting))

# paramslog = {}

# from tqdm import tqdm

# for _req in tqdm(_reqlist):
#     _rt, _ltd, _bst = _req
#     _ltddate = _ltd.split(" ")[0]
#     stkrange = range(int(qqqpxpd.loc[_ltddate]['Low'])-20, int(qqqpxpd.loc[_ltddate]['High'])+20)
#     for _stk in stkrange:
#         _contract = init_qqq_options(_ltddate, _rt, _stk)
        
#         _reqid = app.nextId()
#         paramslog[_reqid] = [_rt, _ltd, _bst, _ltddate, _stk]
#         print("Processing", _reqid, paramslog[_reqid])
#         app.reqHistoricalData(
#             _reqid, 
#             _contract, 
#             _ltd, 
#             _bst[0], 
#             _bst[1], 
#             _bst[2], 
#             1, 
#             1, 
#             False, 
#             [])
        
#         try:
#             get_qqq_bardatapd(app.datapool[_reqid], _reqid)
#         except:
#             print("extract datapd failed", _reqid)
        
#         if _reqid%10==0:
#             time.sleep(10)


# time.sleep(60)

# strdatapool = {}
# for k, v in app.datapool.items():
#     strdatapool[k] = [str(x) for x in v]

# for k, v in strdatapool.items():
#     _suffix = "{}{}{}_{}".format(paramslog[k][3], paramslog[k][0], paramslog[k][4], paramslog[k][2][1])
#     get_qqq_bardatapd(v, k, _suffix, 
#                       savepath="/Users/my/Library/Mobile Documents/com~apple~CloudDocs/env/projects/us_intraday/QQQ/options/", paramslog=paramslog)

    


import requests
import json
import time
import  re
# import pymssql
# import pymysql
import  datetime
import random
from requests.adapters import HTTPAdapter
import pandas as pd
import numpy as np
from functools import reduce
import matplotlib.pyplot as plt

datapath = "./"

import logging
LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
logging.basicConfig(filename=f'{datapath}sina_hourlyupdates.log', level=logging.DEBUG, format=LOG_FORMAT)

def daily_sina_limit_notice():
    page = 0
    while True:
        try:
            page+=1
            referer_url = "http://finance.sina.com.cn/7x24/?tag=0"
            cookie = "UOR=www.baidu.com,tech.sina.com.cn,; SINAGLOBAL=114.84.181.236_1579684610.152568; UM_distinctid=16fcc8a8b704c8-0a1d2def9ca4c6-33365a06-15f900-16fcc8a8b718f1; lxlrttp=1578733570; gr_user_id=2736e487-ee25-4d52-a1eb-c232ac3d58d6; grwng_uid=d762fe92-912b-4ea8-9a24-127a43143ebf; __gads=ID=d79f786106eb99a1:T=1582016329:S=ALNI_MZoErH_0nNZiM3D4E36pqMrbHHOZA; Apache=114.84.181.236_1582267433.457262; ULV=1582626620968:6:4:1:114.84.181.236_1582267433.457262:1582164462661; ZHIBO-SINA-COM-CN=; SUB=_2AkMpBPEzf8NxqwJRmfoWz2_ga4R2zQzEieKfWADoJRMyHRl-yD92qm05tRB6AoTf3EaJ7Bg2UU4l1CDZXUBCzEuJv3mP; SUBP=0033WrSXqPxfM72-Ws9jqgMF55529P9D9WhqhhGsPWdPjar0R99pFT8s"
            headers = {
                "Accept": "*/*",
                "Accept-Encoding": "gzip, deflate",
                "Accept-Language": "zh-CN,zh;q=0.9",
                "Connection": "keep-alive",
                "Cookie": cookie,
                "Host": "zhibo.sina.com.cn",
                "Referer": referer_url,
                "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36"
            }
            base_url = "http://zhibo.sina.com.cn/api/zhibo/feed?callback=jQuery0&page=%s"%page+"&page_size=20&zhibo_id=152&tag_id=0&dire=f&dpc=1&pagesize=20&_=0%20Request%20Method:GET%27"
            data = get_json_data(base_url,headers)
            time.sleep(10)
            print("+", end=' ')
        except Exception as e:
            print('error', e)
            continue

def get_json_data(base_url,headers):
    s = requests.Session()
    s.mount('http://', HTTPAdapter(max_retries=3))
    s.mount('https://', HTTPAdapter(max_retries=3))
    print(time.strftime('%Y-%m-%d %H:%M:%S'))
 
    try:
        response = requests.get(base_url, timeout=5, headers=headers)
        html = response.text
        # print(html)
        html_cl = html[12:-14]
        false = False
        true = True
        null = None
        html_json = eval(html_cl)
        json_str = json.dumps(html_json)
        results = json.loads(json_str)
        data = results['result']['data']['feed']['list']
    except Exception as e:
        print('get_json_str未收录错误类型，请检查网络通断,错误位置：',e)
        time.sleep(5)
        get_json_data(base_url, headers)
    else:
        return data
    
def find_pngurl(data):
    for e in data:
        if '涨停分析' in e['rich_text']:
    #         pngurl=e['multimedia']['img_url'][0]
            return e['multimedia']['img_url'][0]
    return None
        
def get_and_save_png(pngurl, savename):
    import urllib
    urllib.request.urlretrieve(pngurl.replace('\\', ''), "./{}.png".format(savename))

def mail_outlook(filepathlist, subject="Sina Hourly Updates"):
    """
    https://djangocentral.com/sending-emails-with-csv-attachment-using-python/
    https://stackoverflow.com/questions/23171140/how-do-i-send-an-email-with-a-csv-attachment-using-python
    """
    import smtplib
    from email.mime.multipart import MIMEMultipart
    from email.mime.application import MIMEApplication
    from email.mime.text import MIMEText
    from email.mime.image import MIMEImage

    frommail="mydotu@outlook.sg"
    tomail="tan.mengyuan@uobgroup.com"
    password="Tmy07121997"

    msg=MIMEMultipart()
    msg["From"]=frommail
    msg["To"]=tomail
    msg["Subject"]=subject
    body_part = MIMEText("Hi, Attached hourly Sina News...", 'plain')
    
    msg.attach(body_part)
    for fpath in filepathlist:
        with open(fpath, 'rb') as f:
            if ".csv" in fpath:
                _attach = MIMEApplication(f.read(), Name=fpath.split("/")[-1])
            elif ".png" in fpath:
                _attach = MIMEImage(f.read(), Name=fpath.split("/")[-1])
            else:
                pass
            msg.attach(_attach)
        
    server = smtplib.SMTP("smtp-mail.outlook.com", 587)
    # server = smtplib.SMTP("smtp.gmail.com:587")
    check = server.ehlo()
    server.starttls()
    server.login(frommail, password)
    server.sendmail(frommail, tomail, msg.as_string())
    server.quit()
    
def sina_web_common(page):
    referer_url = "http://finance.sina.com.cn/7x24/?tag=0"
    cookie = "UOR=www.baidu.com,tech.sina.com.cn,; SINAGLOBAL=114.84.181.236_1579684610.152568; UM_distinctid=16fcc8a8b704c8-0a1d2def9ca4c6-33365a06-15f900-16fcc8a8b718f1; lxlrttp=1578733570; gr_user_id=2736e487-ee25-4d52-a1eb-c232ac3d58d6; grwng_uid=d762fe92-912b-4ea8-9a24-127a43143ebf; __gads=ID=d79f786106eb99a1:T=1582016329:S=ALNI_MZoErH_0nNZiM3D4E36pqMrbHHOZA; Apache=114.84.181.236_1582267433.457262; ULV=1582626620968:6:4:1:114.84.181.236_1582267433.457262:1582164462661; ZHIBO-SINA-COM-CN=; SUB=_2AkMpBPEzf8NxqwJRmfoWz2_ga4R2zQzEieKfWADoJRMyHRl-yD92qm05tRB6AoTf3EaJ7Bg2UU4l1CDZXUBCzEuJv3mP; SUBP=0033WrSXqPxfM72-Ws9jqgMF55529P9D9WhqhhGsPWdPjar0R99pFT8s"
    headers = {
        "Accept": "*/*",
        "Accept-Encoding": "gzip, deflate",
        "Accept-Language": "zh-CN,zh;q=0.9",
        "Connection": "keep-alive",
        "Cookie": cookie,
        "Host": "zhibo.sina.com.cn",
        "Referer": referer_url,
        "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36"
    }
    base_url = "http://zhibo.sina.com.cn/api/zhibo/feed?callback=jQuery0&page=%s"%page+"&page_size=100&zhibo_id=152&tag_id=0&dire=f&dpc=1&pagesize=20&_=0%20Request%20Method:GET%27"
    return base_url,headers
    
def get_info_from_json(jsonlist):
    _resdict = {
        "id": [],
        "txt": [],
        "time": [],
    }
    _tradable = {
        "market": [],
        "symbol": [],
        "key": [],
        "id": []
    }
    for d in jsonlist:
        _resdict["id"].append(d["id"])
        _resdict["txt"].append(d["rich_text"])
        _resdict["time"].append(d["create_time"])
        try:
            _ext = json.loads(d['ext'])
            if len(_ext['stocks'])>0:
                for s in _ext['stocks']:
                    _tradable['market'].append(s['market'])
                    _tradable['symbol'].append(s['symbol'])
                    _tradable['key'].append(s['key'])
                    _tradable['id'].append(d["id"])
        except Exception as e:
            logging.error(e)
    return _resdict, _tradable

def hourly_updates(ps=[1,2,3,4], send_mail=True):
    respd, tradablepd = get_hourly_datapd(ps=ps)
    respd.to_csv(f"{datapath}sina_news_txt.csv", index=False)
    tradablepd, keycount = process_tradablepd(tradablepd)
    tradablepd.to_csv(f"{datapath}sina_news_tradable.csv", index=False)
    keycount.to_frame().to_csv(f"{datapath}keycount.csv",)
    
    filepathlist = [
        f"{datapath}sina_news_txt.csv",
        f"{datapath}sina_news_tradable.csv",
        f"{datapath}keycount.csv"
    ]
    if send_mail:
        mail_outlook(filepathlist, subject="Sina Updates")
    else:
        return respd, tradablepd, keycount.to_frame()

def get_hourly_datapd(ps=[1,2,3,4]):
    resdicts = []
    tradabledicts = []
    for page in ps:
        base_url, headers = sina_web_common(page)
        jsonlist = get_json_data(base_url, headers)
        try:
            assert len(jsonlist)==100
        except:
            logging.warning("Return records number less than 100...")
        _resdict, _tradable = get_info_from_json(jsonlist)
        resdicts.append(_resdict)
        if len(_tradable['symbol'])==0:
            logging.info("No tradable records...")
        else:
            tradabledicts.append(_tradable)
    respd = pd.concat([pd.DataFrame(x) for x in resdicts]).sort_values("id")
    tradablepd = pd.concat([pd.DataFrame(x) for x in tradabledicts]).sort_values("id")
    return respd[["id", "time", "txt"]], tradablepd[["id", "market", "symbol", "key"]]

def process_tradablepd(tradablepd):
    keycount = tradablepd.groupby("key").count()['id'].sort_values(ascending=False)
#     marketcount = tradablepd.groupby("market").count()['id'].sort_values(ascending=False)
    keycount.name="keycount"
#     marketcount.name="mktcount"
    tradablepd = tradablepd.set_index("key").join(keycount).sort_values("keycount", ascending=False).reset_index()
    tradablepd = tradablepd[['id', 'key', 'keycount', 'market', 'symbol']]
    return tradablepd, keycount
    


from tsapix.utils import *
from tsapix.utils import icloud_froot, env_db_root, code_univ_save_path
from tsapix.utils import save_as_yaml, read_yaml_file


sh_sec_common_headers = {
    'Referer': 'https://www.sse.com.cn/'
}

def update_cn_ticker_universe(
        sh_code_rawf_path = env_db_root+"sh_a_code.xls",
        sz_code_rawf_path = env_db_root+'sz_a_code.xlsx',
):
    # SH Code

    # sh_code_rawf_path = "sh_a_code.xls"

    import requests

    xls_url = """https://query.sse.com.cn/sseQuery/commonExcelDd.do?sqlId=COMMON_SSE_CP_GPJCTPZ_GPLB_GP_L&type=inParams&CSRC_CODE=&STOCK_CODE=&REG_PROVINCE=&STOCK_TYPE=1&COMPANY_STATUS=2,4,5,7,8"""
    xres = requests.get(xls_url, headers={
        'Referer': 'https://www.sse.com.cn/'
    })

    with open(sh_code_rawf_path, 'wb') as f:
        f.write(xres.content)

    sh_code_univlist = [str(x) for x in pd.read_excel(sh_code_rawf_path).iloc[:, 0].values]
    print('Shanghai', len(sh_code_univlist))

    # SZ Code

    # pip install openpyxl --upgrade

    # sz_code_rawf_path = 'sz_a_code.xlsx'

    xlsx_url = """http://www.szse.cn/api/report/ShowReport?SHOWTYPE=xlsx&CATALOGID=1110&TABKEY=tab1&random=0.2520163050277727"""
    xres = requests.get(xlsx_url, headers={
    #     'Referer': 'http://www.szse.cn/'
    })

    with open(sz_code_rawf_path, 'wb') as f:
        f.write(xres.content)

    def sz_code_helper(x):
        if len(x)<6:
            return '0'*(6-len(x))+x
        return x

    sz_code_univlist = [sz_code_helper(str(x)) for x in pd.read_excel(sz_code_rawf_path).iloc[:, 4].values]
    print('Shenzhen', len(sz_code_univlist))

    # CN Code Universe

    a_code_univlist = sh_code_univlist + sz_code_univlist
    print('CN Universe', len(a_code_univlist))

    # Save as .yaml

    # code_univ_save_path = icloud_froot + 'share/' + 'a_code_univ.yaml'
    print('Universe file save path: ', code_univ_save_path)
    save_as_yaml(code_univ_save_path, a_code_univlist)

    # return a_code_univlist


# 上海证券交易所函数 指数代码不同
def get_sh_index_constituents(xindex_code, save_path_root=env_db_root, sh_sec_common_headers=sh_sec_common_headers):
    xurl = f"""http://query.sse.com.cn/commonSoaQuery.do?jsonCallBack=&sqlId=DB_SZZSLB_CFGLB&indexCode={xindex_code}&isPagination=false&_="""
    xres = requests.get(xurl, headers=sh_sec_common_headers)
    xconstituents = [str(x) for x in pd.DataFrame(json.loads(xres.text)['pageHelp']['data'])['securityCode'].values]

    xsave_path = save_path_root + f'constituents_{xindex_code}.yaml'
    save_as_yaml(xsave_path, xconstituents) # only contain live code
    print(len(xconstituents), xsave_path)
    
    xsave_path2 = save_path_root + f'constituentshist_{xindex_code}.yaml'
    xreslist = json.loads(xres.text)['pageHelp']['data']
    try:
        xreslist += read_yaml_file(xsave_path2)
    except FileNotFoundError:
        pass
    save_as_yaml(xsave_path2, xreslist)
    print(len(xreslist), xsave_path2)
    return xsave_path, xconstituents, xsave_path2, xreslist

def get_sh_index_stat(xindex_code, save_path_root=env_db_root, sh_sec_common_headers=sh_sec_common_headers):
    xurl = f"""http://query.sse.com.cn/commonSoaQuery.do?&jsonCallBack=&isPagination=false&sqlId=DB_SZZSLB_QZHYLB&indexCode={xindex_code}&_="""
    xres = requests.get(xurl, headers=sh_sec_common_headers)
    
    xreslist = json.loads(xres.text)['pageHelp']['data']
    xsave_path = save_path_root + f'stat_{xindex_code}.yaml'
    try:
        xreslist += read_yaml_file(xsave_path)
    except FileNotFoundError:
        pass
    save_as_yaml(xsave_path, xreslist)

    print(len(xreslist), xsave_path)
    return xsave_path, xreslist





import os
import sys

from tsapix.utils import read_yaml_file, save_as_yaml, update_yaml
from tsapix.utils import save_npcube, read_npcube
from tsapix.utils import initiate_logger
from tsapix.utils import icloud_froot, env_db_root, cube_warehouse_root, code_univ_save_path, _log_froot

import requests
import json
import numpy as np
import pandas as pd
import multitasking as _multitasking
from tqdm import tqdm
import re
import pandas_market_calendars as mcal

import pandas as _pd
import numpy as _np
import sys as _sys

def empty_df(index=None, cols=[], index_name=None):
    if index is None:
        index = []
    empty = _pd.DataFrame(index=index, data={c:_np.nan for c in cols})
    if index_name:
        empty.index.name = index_name
    return empty


class ProgressBar:
    def __init__(self, iterations, text='completed'):
        self.text = text
        self.iterations = iterations
        self.prog_bar = '[]'
        self.fill_char = '*'
        self.width = 50
        self.__update_amount(0)
        self.elapsed = 1

    def completed(self):
        if self.elapsed > self.iterations:
            self.elapsed = self.iterations
        self.update_iteration(1)
        print('\r' + str(self), end='', file=_sys.stderr)
        _sys.stderr.flush()
        print("", file=_sys.stderr)

    def animate(self, iteration=None):
        if iteration is None:
            self.elapsed += 1
            iteration = self.elapsed
        else:
            self.elapsed += iteration

        print('\r' + str(self), end='', file=_sys.stderr)
        _sys.stderr.flush()
        self.update_iteration()

    def update_iteration(self, val=None):
        val = val if val is not None else self.elapsed / float(self.iterations)
        self.__update_amount(val * 100.0)
        self.prog_bar += f"  {self.elapsed} of {self.iterations} {self.text}"

    def __update_amount(self, new_amount):
        percent_done = int(round((new_amount / 100.0) * 100.0))
        all_full = self.width - 2
        num_hashes = int(round((percent_done / 100.0) * all_full))
        self.prog_bar = '[' + self.fill_char * num_hashes + ' ' * (all_full - num_hashes) + ']'
        pct_place = (len(self.prog_bar) // 2) - len(str(percent_done))
        pct_string = f'{percent_done}%%'
        self.prog_bar = self.prog_bar[0:pct_place] + (pct_string + self.prog_bar[pct_place + len(pct_string):])

    def __str__(self):
        return str(self.prog_bar)




import pandas as pd
import numpy as np
from datetime import datetime
import time
from tqdm import tqdm

import os
# icloud_froot = "/Users/{}/Library/Mobile Documents/com~apple~CloudDocs/".format(os.environ['USER'])
from tsapi.utils import icloud_froot

import talib

# def _color_txt(xstr):
#     return '\033[91m'+xstr+'\033[0m'

# def _black_txt(xstr):
#     return '\033[30m'+xstr+'\033[0m'

######### FORMATTING
def print_infotbl(timestr=['', ''], con1=['', ''], nbys=['', ''], nbysprob=['', ''], _space=20, n=4, rs=2):
    print('- '*int(n*_space/2 + n-1))
    xstr1 = '\033[30m'+'NBProb'+'\033[0m'
    xstr2 = '\033[30m'+'NB'+'\033[0m'
    xstr3 = '\033[30m'+'con1'+'\033[0m'
    print(f"{'Time': ^{_space}}|{xstr1 : ^{_space}}|{xstr2 : ^{_space}}|{xstr3 : ^{_space}}")
    print('- '*int(n*_space/2 + n-1))
    for i in range(rs):
        print(f"{timestr[i] : ^{_space}}|{nbysprob[i] : ^{_space}}|{nbys[i] : ^{_space}}|{con1[i] : ^{_space}}")
        print('- '*int(n*_space/2 + n-1))
#     print('Time'.ljust(_space), '|', 'status_1'.ljust(_space), '|', 'status_0'.ljust(_space), '|')
    return
#########

def add_macd_elements_topd(xpd, pxcol='Adj Close'):
    xpd = xpd.copy(deep=True)
    prefix = '' if pxcol in ['Adj Close', 'Close'] else pxcol + "_"
    xpd[f'{prefix}macd'], xpd[f'{prefix}macd_signal'], xpd[f'{prefix}macd_hist'] = talib.MACD(xpd[pxcol], fastperiod=12, slowperiod=26, signalperiod=9)
    return xpd

def add_macd_elements_topd_paras(xpd, pxcol='Adj Close', fastperiod=12, slowperiod=26, signalperiod=9):
    _suffix = f"_{fastperiod}_{slowperiod}_{signalperiod}"
    xpd = xpd.copy(deep=True)
    prefix = '' if pxcol in ['Adj Close', 'Close'] else pxcol + "_"
    xpd[f'{prefix}macd{_suffix}'], \
        xpd[f'{prefix}macd_signal{_suffix}'], \
            xpd[f'{prefix}macd_hist{_suffix}'] = talib.MACD(xpd[pxcol], 
                fastperiod=fastperiod, slowperiod=slowperiod, signalperiod=signalperiod)
    return xpd

def add_kdj_elements_topd(xpd, hcol='High', lcol='Low', ccol='Close', prefix=None):
    xpd = xpd.copy(deep=True)
    if prefix is None:
        prefix = '' if ccol in ['Adj Close', 'Close'] else ccol + "_"
    xpd[f'{prefix}k'], xpd[f'{prefix}d'] = talib.STOCH(
        xpd[hcol],
        xpd[lcol],
        xpd[ccol],
        fastk_period=9,
        slowk_period=3,
        slowk_matype=0,
        slowd_period=3,
        slowd_matype=0
    )
    xpd[f'{prefix}j'] = 3 * xpd[f'{prefix}k'] - 2 * xpd[f'{prefix}d']
    return xpd

def add_kdj_elements_topd_paras(xpd, hcol='High', lcol='Low', ccol='Close', prefix=None, 
                                fastk_period=9, slowk_period=3, slowd_period=3):
    _suffix = f"_{fastk_period}_{slowk_period}_{slowd_period}"
    xpd = xpd.copy(deep=True)
    if prefix is None:
        prefix = '' if ccol in ['Adj Close', 'Close'] else ccol + "_"
    _k, _d, _j = f'{prefix}k{_suffix}', f'{prefix}d{_suffix}', f'{prefix}j{_suffix}'
    xpd[_k], xpd[_d] = talib.STOCH(
        xpd[hcol],
        xpd[lcol],
        xpd[ccol],
        fastk_period=9,
        slowk_period=3,
        slowk_matype=0,
        slowd_period=3,
        slowd_matype=0
    )
    xpd[_j] = 3 * xpd[_k] - 2 * xpd[_d]
    return xpd

def add_rolling_rank(xpd, xcols=['macd', 'j'], rws=[30, 60], scale=True):
    xpd = xpd.copy(deep=True)
    for rw in rws:
        for c in tqdm(xcols):
            _colname = f"{c}_rel{rw}"
            xpd[_colname] = xpd[c].rolling(rw).apply(lambda x: np.argsort(np.argsort(x)).values[-1])
            if scale:
                xpd[_colname] = xpd[_colname] / (rw - 1)
    return xpd
    
def print_trading_bar():
    from datetime import datetime, timedelta
    
    us_now = datetime.now() - timedelta(minutes=13*60)
    trade_date = us_now.strftime("%Y-%m-%d")
    trade_starttime = trade_date + " 09:30"
    trade_starttime = datetime.strptime(trade_starttime, "%Y-%m-%d %H:%M")
    trade_endtime = trade_date + " 16:00"
    trade_endtime = datetime.strptime(trade_endtime, "%Y-%m-%d %H:%M")
    
    total_secs = (trade_endtime - trade_starttime).seconds
    gone_secs =(us_now - trade_starttime).seconds
    
    print("Trading Bar: {:.2f}% ({}/{} secs)".format(
        100 * (gone_secs / total_secs),
        gone_secs,
        total_secs
    ))

#############################################################
# trading interface WEBULL
import pyautogui

positions_dict = {
    'symbol': (276, 375),
    'side': {
        'buy': (460, 187),
        'sell': (585, 190),
    },
    'placeorder_market': (474, 353),
    'placeorder_limit': (512, 416),
    'num': (559, 251),

}

positions_dict_options = {
    'symbol': (276, 375),
    'side': {
        'buy': (460, 253),
        'sell': (585, 249),
    },
    'placeorder_market': (569, 478),
#     'placeorder_limit': (543, 542),
    'num': (550, 313),

}

_order_gap = 5


def place_market_order(side='buy', num=1, instrument='equity'):
    if instrument=='equity':
        xydict = positions_dict.copy()
    elif instrument=='options':
        xydict = positions_dict_options.copy()
#     pyautogui.click(xydict['symbol'])
    pyautogui.click(xydict['side'][side])
    if num>1:
        modify_shares_num(num=num, positions_dict=xydict)
    pyautogui.click(xydict['placeorder_market'])
    # pyautogui.click(xydict['placeorder_limit']) #### delete
    _order_timstr = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(side.upper(), 'market order placed at ', _order_timstr, "(Instrument: ", instrument, ")")
    return {
        '_order_timstr': _order_timstr,
        'side': side,
    }

def modify_shares_num(num=1, positions_dict=positions_dict):
    pyautogui.click(positions_dict['num'])
    pyautogui.press('backspace')
    pyautogui.write(str(num))
    pyautogui.press('enter')


def live():
    return {
        '_data_timstr': 0,
        'status_1': False,
        'status_0': False,
    }

class LiveTrader:

    def __init__(self, 
                 indfunc=live, 
                 trading_indstr='status_0',
                 inipos=0,
                 instrument='equity',
                ):
        self.position = inipos
        self.order_records = {}
        self.indfunc = indfunc
        self.trading_indstr = trading_indstr
        self.instrument = instrument

        self.metrics = {
            'buy': 0,
            'sell': 0,
        } # record some metrics

        self.cur_date = datetime.now().strftime("%Y-%m-%d")
        self.order_records_savedir = icloud_froot + 'env/projects/tsla_1yr/trading_log/'
        self.order_records_savepath = f"{self.order_records_savedir}/{self.cur_date}.csv"

    def log_records(self):
        try:
            pd.DataFrame(self.order_records).T.to_csv(self.order_records_savepath)
        except Exception as e:
            print('Order Log Failed, ', e)

    def log_trading_metrics(self):
        try:
            print('Trade times: ', 'buy', self.metrics['buy'], '; sell', self.metrics['sell'], '; sum', self.metrics['buy'] + self.metrics['sell'])
        except Exception as e:
            print('Metrics Log Failed, ', e)

    def pure_long_trading(self):
        try:
            xdict = self.indfunc()
        except Exception as e:
            print("Error updating... ", e)
            return

        trading_indstr = self.trading_indstr
        # trading_indstr = 'random'
        print(trading_indstr)

        if trading_indstr == 'random':
            xdict[trading_indstr] = np.random.random() > 0.5

        # # place buy order
        print(xdict[trading_indstr], self.position, xdict[trading_indstr]==False, str(xdict[trading_indstr])=='False', int(xdict[trading_indstr])==0)
        if np.all([
            str(xdict[trading_indstr])=='False' or int(xdict[trading_indstr])==0,
            self.position==0
        ]): 
            _odt = place_market_order(side='buy')
            self.position+=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }
            self.metrics['buy'] += 1

        # # place sell order
        if np.all([
            str(xdict[trading_indstr])=='True' or int(xdict[trading_indstr])==1,
            self.position>0
        ]):
            while self.position>0:
                _odt = place_market_order(side='sell')
                self.position-=1
                self.order_records[_odt['_order_timstr']] = {
                    'side': _odt['side'],
                    'pos': self.position,
                }

                self.metrics['sell'] += 1

        print("Current Position: ", self.position)
        self.log_records()
        self.log_trading_metrics()

    def long_short_trading(self, ):
        try:
            xdict = self.indfunc()
        except Exception as e:
            print("Error updating... ", e)
            return

        trading_indstr = self.trading_indstr
        # trading_indstr = 'random'
        print(trading_indstr)

        if trading_indstr == 'random':
            xdict[trading_indstr] = np.random.random() > 0.5

        # # place buy order
        print(xdict[trading_indstr], self.position, xdict[trading_indstr]==False, str(xdict[trading_indstr])=='False', int(xdict[trading_indstr])==0)
        if np.all([
            str(xdict[trading_indstr])=='False' or int(xdict[trading_indstr])==0,
            self.position==0
        ]): 
            _odt = place_market_order(side='buy', instrument=self.instrument)
            self.position+=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }
            self.metrics['buy'] += 1

        if np.all([
            str(xdict[trading_indstr])=='False' or int(xdict[trading_indstr])==0,
            self.position==-1
        ]): 
            for i in range(2):
                _odt = place_market_order(side='buy', num=1, instrument=self.instrument)
                self.position+=1
                self.order_records[_odt['_order_timstr']] = {
                    'side': _odt['side'],
                    'pos': self.position,
                }
                self.metrics['buy'] += 1
                time.sleep(_order_gap) # in case order not filled

        # # place sell order
        if np.all([
            str(xdict[trading_indstr])=='True' or int(xdict[trading_indstr])==1,
            self.position==0
        ]):
            _odt = place_market_order(side='sell', instrument=self.instrument)
            self.position-=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }

            self.metrics['sell'] += 1

        if np.all([
            str(xdict[trading_indstr])=='True' or int(xdict[trading_indstr])==1,
            self.position==1
        ]):
            for i in range(2):
                _odt = place_market_order(side='sell', num=1, instrument=self.instrument)
                self.position-=1
                self.order_records[_odt['_order_timstr']] = {
                    'side': _odt['side'],
                    'pos': self.position,
                }

                self.metrics['sell'] += 1
                time.sleep(_order_gap)

        

        print("Current Position: ", self.position)
        self.log_records()
        self.log_trading_metrics()

    def pure_short_trading(self):
        try:
            xdict = self.indfunc()
        except Exception as e:
            print("Error updating... ", e)
            return

        trading_indstr = self.trading_indstr
        # trading_indstr = 'random'
        print(trading_indstr)

        if trading_indstr == 'random':
            xdict[trading_indstr] = np.random.random() > 0.5

        # # place buy order
        print(xdict[trading_indstr], self.position, xdict[trading_indstr]==False, str(xdict[trading_indstr])=='False')
        if np.all([
            str(xdict[trading_indstr])=='False' or int(xdict[trading_indstr])==0,
            self.position==-1
        ]): 
            _odt = place_market_order(side='buy', instrument=self.instrument)
            self.position+=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }
            self.metrics['buy'] += 1

        # # place sell order
        if np.all([
            str(xdict[trading_indstr])=='True' or int(xdict[trading_indstr])==1,
            self.position==0
        ]):
            _odt = place_market_order(side='sell', instrument=self.instrument)
            self.position-=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }

            self.metrics['sell'] += 1

        print("Current Position: ", self.position)
        self.log_records()
        self.log_trading_metrics()
        return
    
    def pure_short_trading_puts(self):
        try:
            xdict = self.indfunc()
        except Exception as e:
            print("Error updating... ", e)
            return

        trading_indstr = self.trading_indstr
        # trading_indstr = 'random'
        print(trading_indstr)

        if trading_indstr == 'random':
            xdict[trading_indstr] = np.random.random() > 0.5

        # # place sell order
        print(xdict[trading_indstr], self.position, xdict[trading_indstr]==False, str(xdict[trading_indstr])=='False')
        if np.all([
            str(xdict[trading_indstr])=='False' or int(xdict[trading_indstr])==0,
            self.position==1
        ]): 
            _odt = place_market_order(side='sell', instrument=self.instrument)
            self.position-=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }
            self.metrics['sell'] += 1

        # # place buy order
        if np.all([
            str(xdict[trading_indstr])=='True' or int(xdict[trading_indstr])==1,
            self.position==0
        ]):
            _odt = place_market_order(side='buy', instrument=self.instrument)
            self.position+=1
            self.order_records[_odt['_order_timstr']] = {
                'side': _odt['side'],
                'pos': self.position,
            }

            self.metrics['buy'] += 1

        print("Current Position: ", self.position)
        self.log_records()
        self.log_trading_metrics()
        return




import requests
from tsapi.utils import icloud_froot, save_as_yaml
import os

def get_nasdaq100_constituents(savepath=icloud_froot + "share/"):
    # headers={"user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.107 Safari/537.36"}
    headers={
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36",
        "Accept": "application/json, text/plain, */*",
        "Accept-Language": "en-US,en;q=0.9",
        "Origin": "https://www.nasdaq.com",
        "Referer": "https://www.nasdaq.com/",
    }
    res=requests.get("https://api.nasdaq.com/api/quote/list-type/nasdaq100",headers=headers)
    main_data=res.json()['data']['data']['rows']

    nasdaq100_tickerpool = []
    for i in range(len(main_data)):
        print(main_data[i]['symbol'], end=" ")
        nasdaq100_tickerpool.append(main_data[i]['symbol'])

    if savepath is not None:
        save_as_yaml(os.path.join(savepath, "nasdaq100_tickerpool.yaml"), nasdaq100_tickerpool)
    return nasdaq100_tickerpool


from tsapix.utils import icloud_froot, cube_warehouse_root

from tsapix.spider.base import load_cube_n_axes, get_dtbk, load_maps
from tsapix.utils import nan_argsort

from datetime import datetime, timedelta
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm

cuberoot=cube_warehouse_root

dtformatdict={
            'd': "%Y-%m-%d",
            'w': "%Y-%m-%d",
            'm': "%Y-%m-%d",
            '5min': "%Y-%m-%d %H:%M",
            '15min': "%Y-%m-%d %H:%M",
            '30min': "%Y-%m-%d %H:%M",
            '60min': "%Y-%m-%d %H:%M",
        }

maps = load_maps()
code_name_mapdict = maps['stock_meta']

# def maps():
#     return tsapi.base.load_maps()


def load_histpx_cubes(
    fts=[
        ('5min', 'bfq'),
        ('d', 'bfq'),
        ('d', 'hfq'),
        ('d', 'qfq'),
    ],
    cuberoot=cuberoot,
    dtformatdict=dtformatdict,
):
    """ Load histpx cube and axes.

    fts: list of tuple (frequency, px type)
    cuberoot: string, cube file save root
    dtformatdict: dict, indicate time format for different frequency
    """
    histpx = {}
    for frq, tp in fts:
        _k = f'{frq}_{tp}'
        print("Loading ", _k, "... ", end=" ")

        histpx[_k]={}
        cube_dict = load_cube_n_axes(f'histpx_{_k}', froot=cuberoot, timeformat=dtformatdict[frq])
        _cube = cube_dict['cube']
        _axs = cube_dict['axes']
        histpx[_k]['cube'] = _cube
        histpx[_k]['axes'] = _axs

        for fld in _axs[0]:
            print(fld, end="; ")
            histpx[_k][fld] = pd.DataFrame(
                                    _cube[_axs[0].index(fld), :, :], 
                                    index=_axs[1], 
                                    columns=_axs[2]).ffill()
        print()
    return histpx

def get_boll_lines(
    pxpd,
    _w=20,
    _nstd=2,
):
    """ Calculate BOLL elements.

    pxpd: DataFrame, price for calculate BOLL
    _w: int, rolling window
    _nstd: int / float, times of std add on / subtruct from mean
    
    """
    _mean = pxpd.copy(deep=True).rolling(_w).mean()
    _std = pxpd.copy(deep=True).rolling(_w).std()

    _upper_bond = _mean + _nstd * _std
    _lower_bond = _mean - _nstd * _std
    return {
        'px': pxpd,
        'm': _mean,
        's': _std,
        'u': _upper_bond,
        'l': _lower_bond,
        '%b': (pxpd - _lower_bond) / (_upper_bond - _lower_bond)
    }

def get_rchgs(
    chgpd,
):
    """ Calculate relative percentage change strength.
    
    chgpd: DataFrame, price percent change
    """
    pchgspd = pd.DataFrame(
        chgpd.apply(lambda x: nan_argsort(x, ascending=False), axis=1).to_list(), 
        index=chgpd.index,
        columns=chgpd.columns
    )

    _d = pchgspd.max(axis=1).values
    pchgspd_01 = pchgspd.apply(lambda x: x/_d)
    return {
        'rank': pchgspd.copy(deep=True),
        '01': pchgspd_01.copy(deep=True)
    }

def daily_intraday_toprchg(histpx, ft='5min_bfq', topn=20):
    """ Times ranked top; return daily frequency.

    histpx: dict, output of load_histpx_cubes
    ft: string, indicate frequency and price type
    topn: int, threshold for consideration as top
    """
    chgpd = 100 * ((histpx[ft]['High'] / histpx[ft]['Close'].shift(1)) - 1)
    rchgs = get_rchgs(chgpd)
    dbk = get_dtbk(chgpd.index[0].strftime("%Y-%m-%d"), chgpd.index[-1].strftime("%Y-%m-%d"), 'd')
    toprchg = dbk.join((rchgs['rank'] < topn).resample('d').sum())
    return toprchg

def add_rel_rel01(xdict, _ks):
    for k in _ks:
        xdict[f"{k}_rel"] = pd.DataFrame(
            xdict[k].copy(deep=True).apply(lambda x: nan_argsort(x, ascending=False), axis=1).to_list(), 
            index=xdict[k].index,
            columns=xdict[k].columns
        )
        _d = xdict[f"{k}_rel"].max(axis=1).values
        xdict[f"{k}_rel01"] = xdict[f"{k}_rel"].apply(lambda x: x/_d)
    return xdict

def toprchg_indicators(toprchg):
    """ Indicators based on toprchg.

    toprchg: DataFrame, output of daily_intraday_toprchg
    
    """
    res = {
        'v1': (toprchg>0).rolling(5).sum()>3,
        'v2': (toprchg>0).rolling(20).mean(),
        'v3': toprchg.rolling(20).sum(),
    }

    res = add_rel_rel01(res, _ks=['v2', 'v3'])
    return res


import matplotlib.pyplot as plt
import tsapi
import talib
import numpy as np
import streamlit as st

from matplotlib import font_manager

fontP = font_manager.FontProperties()
fontP.set_family('SimHei')
fontP.set_size(14)


def obs_names_at_specific_mth(_names=[], _ym="xxxx-xx",
                              xobj=None, 
                              code_name_mapdict=None,
                              xpd_ma20=None,
                              xpd_ma5=None,
                              b_low=None,
                              b_up=None,
                              xpd=None,
                              obs_step=20,
                              additional_info=[],
                              savepath=None,
                              show=True
                              ):
# _names = ['603099',]
# _ym = '2024-01'
    histpx_d_qfq_axes = xobj['axes']
    pxpddict = xobj

    _dates = [x for x in histpx_d_qfq_axes['1str'] if _ym in x]
    _start_date_id = histpx_d_qfq_axes['1str'].index(_dates[0])
    _end_date_id = histpx_d_qfq_axes['1str'].index(_dates[-1])

    _plot_range = slice(_start_date_id-obs_step, _end_date_id+obs_step)
    _fig_size = (16, 5)
    _ber_length = 0.3


    xxid = 0
    for _plot_ticker in _names[:]:
        try:
            print(_plot_ticker, code_name_mapdict[_plot_ticker]['name'], end=";")
        except:
            pass
        if len(additional_info)>0:
            for adxsuit in additional_info:
                if type(adxsuit) in [list, tuple]:
                    if len(adxsuit)==2:
                        adx, _prefix = adxsuit
                        _suffix = ''
                    else:
                        adx, _prefix, _suffix = adxsuit
                else:
                    adx = adxsuit
                    _prefix, _suffix = '', ''
                try:
                    print(_prefix, adx[_plot_ticker], _suffix, end=";")
                except:
                    pass
        
        # _loc_id = histpx_d_qfq_axes[2].index(_plot_ticker)

        fig, axs = plt.subplots(figsize=_fig_size)
        xs = [x.strftime("%Y-%m-%d") for x in histpx_d_qfq_axes[1][_plot_range]]
        xvs = range(len(xs))
        xticks = range(-len(xs), 0)
        axs.set_xticklabels(xticks)
        
        _plt_open = pxpddict['Open'][_plot_ticker].iloc[_plot_range]
        _plt_high = pxpddict['High'][_plot_ticker].iloc[_plot_range]
        _plt_low = pxpddict['Low'][_plot_ticker].iloc[_plot_range]
        _plt_close = pxpddict['Close'][_plot_ticker].iloc[_plot_range]
        _px_colors = np.where(_plt_close>_plt_open, 'firebrick', 'darkslateblue')
        
        _tr = pxpddict['TurnoverRate'][_plot_ticker].iloc[_plot_range]
        
        try:
            axs.vlines(
                xs, 
                list(_plt_low.values), 
                list(_plt_high.values), 
                color=_px_colors,
                label=code_name_mapdict[_plot_ticker]['name'],
            )
        except:
            axs.vlines(
                    xs, 
                    list(_plt_low.values), 
                    list(_plt_high.values), 
                    color=_px_colors,
                )
        # plt.legend(loc=0, prop=fontP)
        axs.hlines(
            _plt_open, 
            np.array(xvs) - _ber_length, 
            xvs,
            color=_px_colors,
        )
        axs.hlines(
            _plt_close, 
            xvs, 
            np.array(xvs) + _ber_length,
            color=_px_colors,
        )
        
        try:
            axs.plot(xs, xpd_ma20[_plot_ticker].iloc[_plot_range].values, color='orange')
        #     axs.plot(xs, xpd_ma10[_plot_ticker].iloc[_plot_range].values, color='grey')
            axs.plot(xs, xpd_ma5[_plot_ticker].iloc[_plot_range].values, color='grey')
            
            axs.plot(xs, b_up[_plot_ticker].iloc[_plot_range].values, color='blue', alpha=0.5)
            axs.plot(xs, b_low[_plot_ticker].iloc[_plot_range].values, color='purple', alpha=0.5)
        except:
            pass

        axs_0 = axs.twinx()
        axs_0.vlines(obs_step, 0, 1, linestyles='--', colors='black')
        axs_0.vlines(len(xs) - obs_step, 0, 1, linestyles='--', colors='black')

        if savepath is not None:
            try:
                fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}_{code_name_mapdict[_plot_ticker]['name']}.png", bbox_inches='tight')
            except:
                fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}.png", bbox_inches='tight')
        
        fig, axs = plt.subplots(figsize=(16, 3))
        l1pd, l2pd, lhistpd = talib.MACD(xpd[_plot_ticker], fastperiod=12, slowperiod=26, signalperiod=9)
        axs.plot(xvs, l1pd.iloc[_plot_range], color='orange')
        axs.plot(xvs, l2pd.iloc[_plot_range], color='blue')
        axs.bar(xvs, lhistpd.iloc[_plot_range], color='blue')

        if savepath is not None:
            fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}_macd.png", bbox_inches='tight')
        
        if show:
            plt.show()

        xxid+=1

def obs_names_at_specific_mth_v2(_names=[], _ym="xxxx-xx",
                              xobj=None, 
                              code_name_mapdict=None,
                              xpd_ma20=None,
                              xpd_ma5=None,
                              b_low=None,
                              b_up=None,
                              xpd=None,
                              obs_step=20,
                              additional_info=[],
                              savepath=None,
                              show=True,
                              alertedindo=None
                              ):
# _names = ['603099',]
# _ym = '2024-01'
    histpx_d_qfq_axes = xobj['axes']
    pxpddict = xobj

    _dates = [x for x in histpx_d_qfq_axes['1str'] if _ym in x]
    _start_date_id = histpx_d_qfq_axes['1str'].index(_dates[0])
    _end_date_id = histpx_d_qfq_axes['1str'].index(_dates[-1])

    _plot_range = slice(_start_date_id-obs_step, _end_date_id+obs_step)
    _fig_size = (16, 5+2+2)
    _ber_length = 0.3


    xxid = 0
    for _plot_ticker in _names[:]:
        try:
            print(_plot_ticker, code_name_mapdict[_plot_ticker]['name'], end=";")
        except:
            pass
        if len(additional_info)>0:
            for adxsuit in additional_info:
                if type(adxsuit) in [list, tuple]:
                    if len(adxsuit)==2:
                        adx, _prefix = adxsuit
                        _suffix = ''
                    else:
                        adx, _prefix, _suffix = adxsuit
                else:
                    adx = adxsuit
                    _prefix, _suffix = '', ''
                try:
                    print(_prefix, adx[_plot_ticker], _suffix, end=";")
                except:
                    pass
        
        # _loc_id = histpx_d_qfq_axes[2].index(_plot_ticker)

        fig, axs = plt.subplots(nrows=3, ncols=1, height_ratios=[5, 2, 2], figsize=_fig_size, sharex=True, gridspec_kw=dict(hspace=0))
        # axs[0].plot(range(10), range(10))

        # fig, axs = plt.subplots(figsize=_fig_size)
        xs = [x.strftime("%Y-%m-%d") for x in histpx_d_qfq_axes[1][_plot_range]]
        xvs = range(len(xs))
        xticks = range(-len(xs), 0)
        axs[2].set_xticklabels(xticks)
        
        _plt_open = pxpddict['Open'][_plot_ticker].iloc[_plot_range]
        _plt_high = pxpddict['High'][_plot_ticker].iloc[_plot_range]
        _plt_low = pxpddict['Low'][_plot_ticker].iloc[_plot_range]
        _plt_close = pxpddict['Close'][_plot_ticker].iloc[_plot_range]
        _px_colors = np.where(_plt_close>_plt_open, 'firebrick', 'darkslateblue')
        
        _tr = pxpddict['TurnoverRate'][_plot_ticker].iloc[_plot_range]
        
        try:
            axs[0].vlines(
                xs, 
                list(_plt_low.values), 
                list(_plt_high.values), 
                color=_px_colors,
                label=code_name_mapdict[_plot_ticker]['name'],
            )
        except:
            axs[0].vlines(
                    xs, 
                    list(_plt_low.values), 
                    list(_plt_high.values), 
                    color=_px_colors,
                )
        # plt.legend(loc=0, prop=fontP)
        axs[0].hlines(
            _plt_open, 
            np.array(xvs) - _ber_length, 
            xvs,
            color=_px_colors,
        )
        axs[0].hlines(
            _plt_close, 
            xvs, 
            np.array(xvs) + _ber_length,
            color=_px_colors,
        )
        
        try:
            axs[0].plot(xs, xpd_ma20[_plot_ticker].iloc[_plot_range].values, color='orange')
        #     axs.plot(xs, xpd_ma10[_plot_ticker].iloc[_plot_range].values, color='grey')
            axs[0].plot(xs, xpd_ma5[_plot_ticker].iloc[_plot_range].values, color='grey')
            
            axs[0].plot(xs, b_up[_plot_ticker].iloc[_plot_range].values, color='blue', alpha=0.5)
            axs[0].plot(xs, b_low[_plot_ticker].iloc[_plot_range].values, color='purple', alpha=0.5)
        except:
            pass

        _plt_ema12 = pxpddict['Close'][_plot_ticker].ewm(span=12, adjust=False).mean()
        _plt_ema26 = pxpddict['Close'][_plot_ticker].ewm(span=26, adjust=False).mean()
        _plt_ema60 = pxpddict['Close'][_plot_ticker].ewm(span=60, adjust=False).mean()
        _plt_ema130 = pxpddict['Close'][_plot_ticker].ewm(span=130, adjust=False).mean()
        _plt_ema260 = pxpddict['Close'][_plot_ticker].ewm(span=260, adjust=False).mean()
        
        axs[0].plot(xvs, _plt_ema12.iloc[_plot_range], color='red', alpha=0.2)
        axs[0].plot(xvs, _plt_ema26.iloc[_plot_range], color='blue', alpha=0.2)
        axs[0].plot(xvs, _plt_ema60.iloc[_plot_range], color='orange', alpha=0.2)
        axs[0].plot(xvs, _plt_ema130.iloc[_plot_range], color='green', alpha=0.2)
        axs[0].plot(xvs, _plt_ema260.iloc[_plot_range], color='purple', alpha=0.2)

        if alertedindo is not None:
            _xs = [x for x in xs if int(x.replace("-", "")) in alertedindo.index]
            aids = [i for i in range(len(_xs)) if alertedindo[_plot_ticker].loc[int(_xs[i].replace("-", ""))]]

            axs[0].scatter([xs[i] for i in aids], [pxpddict['Close'][_plot_ticker].iloc[_plot_range].iloc[i] for i in aids], c='red', alpha=0.5, s=100)

        axs_0 = axs[0].twinx()
        axs_0.vlines(obs_step, 0, 1, linestyles='--', colors='black')
        axs_0.vlines(len(xs) - obs_step, 0, 1, linestyles='--', colors='black')

        _plt_volume = pxpddict['TurnoverRate'][_plot_ticker].iloc[_plot_range]
        _plt_pctchg = pxpddict['Close'][_plot_ticker].pct_change().iloc[_plot_range]
        volcolors = np.where(_plt_pctchg>0, 'red', 'green')
        axs[1].bar(xvs, _plt_volume, color=volcolors)


        # fig, axs = plt.subplots(figsize=(16, 3))
        l1pd, l2pd, lhistpd = talib.MACD(xpd[_plot_ticker], fastperiod=12, slowperiod=26, signalperiod=9)
        axs[2].plot(xvs, l1pd.iloc[_plot_range], color='orange')
        axs[2].plot(xvs, l2pd.iloc[_plot_range], color='blue')
        axs[2].bar(xvs, lhistpd.iloc[_plot_range], color='blue')

        import pandas as pd
        pd.DataFrame().ewm(span=12, adjust=False)

        plt.show(fig)

        xxid+=1


def obs_names_at_specific_mth_streamlit(_names=[], _ym="xxxx-xx",
                              xobj=None, 
                              code_name_mapdict=None,
                              xpd_ma20=None,
                              xpd_ma5=None,
                              b_low=None,
                              b_up=None,
                              xpd=None,
                              obs_step=20,
                              additional_info=[],
                              savepath=None,
                              show=True,
                              alertedindo=None
                              ):
# _names = ['603099',]
# _ym = '2024-01'
    histpx_d_qfq_axes = xobj['axes']
    pxpddict = xobj

    _dates = [x for x in histpx_d_qfq_axes['1str'] if _ym in x]
    _start_date_id = histpx_d_qfq_axes['1str'].index(_dates[0])
    _end_date_id = histpx_d_qfq_axes['1str'].index(_dates[-1])

    _plot_range = slice(_start_date_id-obs_step, _end_date_id+obs_step)
    _fig_size = (16, 5)
    _ber_length = 0.3


    xxid = 0
    for _plot_ticker in _names[:]:
        try:
            st.write(_plot_ticker, code_name_mapdict[_plot_ticker]['name'], end=";")
        except:
            pass
        if len(additional_info)>0:
            for adxsuit in additional_info:
                if type(adxsuit) in [list, tuple]:
                    if len(adxsuit)==2:
                        adx, _prefix = adxsuit
                        _suffix = ''
                    else:
                        adx, _prefix, _suffix = adxsuit
                else:
                    adx = adxsuit
                    _prefix, _suffix = '', ''
                try:
                    st.write(_prefix, adx[_plot_ticker], _suffix, end=";")
                except:
                    pass
        
        # _loc_id = histpx_d_qfq_axes[2].index(_plot_ticker)

        fig, axs = plt.subplots(figsize=_fig_size)
        xs = [x.strftime("%Y-%m-%d") for x in histpx_d_qfq_axes[1][_plot_range]]
        xvs = range(len(xs))
        xticks = range(-len(xs), 0)
        axs.set_xticklabels(xticks)
        
        _plt_open = pxpddict['Open'][_plot_ticker].iloc[_plot_range]
        _plt_high = pxpddict['High'][_plot_ticker].iloc[_plot_range]
        _plt_low = pxpddict['Low'][_plot_ticker].iloc[_plot_range]
        _plt_close = pxpddict['Close'][_plot_ticker].iloc[_plot_range]
        _px_colors = np.where(_plt_close>_plt_open, 'firebrick', 'darkslateblue')
        
        _tr = pxpddict['TurnoverRate'][_plot_ticker].iloc[_plot_range]
        
        try:
            axs.vlines(
                xs, 
                list(_plt_low.values), 
                list(_plt_high.values), 
                color=_px_colors,
                label=code_name_mapdict[_plot_ticker]['name'],
            )
        except:
            axs.vlines(
                    xs, 
                    list(_plt_low.values), 
                    list(_plt_high.values), 
                    color=_px_colors,
                )
        # plt.legend(loc=0, prop=fontP)
        axs.hlines(
            _plt_open, 
            np.array(xvs) - _ber_length, 
            xvs,
            color=_px_colors,
        )
        axs.hlines(
            _plt_close, 
            xvs, 
            np.array(xvs) + _ber_length,
            color=_px_colors,
        )
        
        try:
            axs.plot(xs, xpd_ma20[_plot_ticker].iloc[_plot_range].values, color='orange')
        #     axs.plot(xs, xpd_ma10[_plot_ticker].iloc[_plot_range].values, color='grey')
            axs.plot(xs, xpd_ma5[_plot_ticker].iloc[_plot_range].values, color='grey')
            
            axs.plot(xs, b_up[_plot_ticker].iloc[_plot_range].values, color='blue', alpha=0.5)
            axs.plot(xs, b_low[_plot_ticker].iloc[_plot_range].values, color='purple', alpha=0.5)
        except:
            pass

        if alertedindo is not None:
            # st.table(alertedindo[_plot_ticker].iloc[-len(xs):])
            # aids = [i for i in range(len(xs)) if alertedindo[_plot_ticker].iloc[-len(xs):].iloc[i]]
            _xs = [x for x in xs if int(x.replace("-", "")) in alertedindo.index]
            aids = [i for i in range(len(_xs)) if alertedindo[_plot_ticker].loc[int(_xs[i].replace("-", ""))]]
            # def _get_alert_state(x):
            #     try:
            #         if alertedindo.loc[int(x.replace("-", ""))][_plot_ticker]:
            #             return xs.index(x)
            #     except:
            #         pass
            # aids = [_get_alert_state(x) for x in xs]
            # st.write(aids)
            axs.scatter([xs[i] for i in aids], [pxpddict['Close'][_plot_ticker].iloc[_plot_range].iloc[i] for i in aids], c='red', alpha=0.5, s=100)

        axs_0 = axs.twinx()
        axs_0.vlines(obs_step, 0, 1, linestyles='--', colors='black')
        axs_0.vlines(len(xs) - obs_step, 0, 1, linestyles='--', colors='black')

        st.pyplot(fig, use_container_width=False)

        if savepath is not None:
            try:
                fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}_{code_name_mapdict[_plot_ticker]['name']}.png", bbox_inches='tight')
            except:
                fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}.png", bbox_inches='tight')
        
        fig, axs = plt.subplots(figsize=(16, 3))
        l1pd, l2pd, lhistpd = talib.MACD(xpd[_plot_ticker], fastperiod=12, slowperiod=26, signalperiod=9)
        axs.plot(xvs, l1pd.iloc[_plot_range], color='orange')
        axs.plot(xvs, l2pd.iloc[_plot_range], color='blue')
        axs.bar(xvs, lhistpd.iloc[_plot_range], color='blue')

        if savepath is not None:
            fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}_macd.png", bbox_inches='tight')
        
        # if show:
            # plt.show()
        st.pyplot(fig, use_container_width=False)

        xxid+=1


def obs_names_at_specific_mth_streamlit_v2(_names=[], _ym="xxxx-xx",
                              xobj=None, 
                              code_name_mapdict=None,
                              xpd_ma20=None,
                              xpd_ma5=None,
                              b_low=None,
                              b_up=None,
                              xpd=None,
                              obs_step=20,
                              additional_info=[],
                              savepath=None,
                              show=True,
                              alertedindo=None
                              ):
# _names = ['603099',]
# _ym = '2024-01'
    histpx_d_qfq_axes = xobj['axes']
    pxpddict = xobj

    _dates = [x for x in histpx_d_qfq_axes['1str'] if _ym in x]
    _start_date_id = histpx_d_qfq_axes['1str'].index(_dates[0])
    _end_date_id = histpx_d_qfq_axes['1str'].index(_dates[-1])

    _plot_range = slice(_start_date_id-obs_step, _end_date_id+obs_step)
    _fig_size = (16, 5+2+2)
    _ber_length = 0.3


    xxid = 0
    for _plot_ticker in _names[:]:
        try:
            st.write(_plot_ticker, code_name_mapdict[_plot_ticker]['name'], end=";")
        except:
            pass
        if len(additional_info)>0:
            for adxsuit in additional_info:
                if type(adxsuit) in [list, tuple]:
                    if len(adxsuit)==2:
                        adx, _prefix = adxsuit
                        _suffix = ''
                    else:
                        adx, _prefix, _suffix = adxsuit
                else:
                    adx = adxsuit
                    _prefix, _suffix = '', ''
                try:
                    st.write(_prefix, adx[_plot_ticker], _suffix, end=";")
                except:
                    pass
        
        # _loc_id = histpx_d_qfq_axes[2].index(_plot_ticker)

        fig, axs = plt.subplots(nrows=3, ncols=1, height_ratios=[5, 2, 2], figsize=_fig_size, sharex=True, gridspec_kw=dict(hspace=0))
        # axs[0].plot(range(10), range(10))

        # fig, axs = plt.subplots(figsize=_fig_size)
        xs = [x.strftime("%Y-%m-%d") for x in histpx_d_qfq_axes[1][_plot_range]]
        xvs = range(len(xs))
        xticks = range(-len(xs), 0)
        axs[2].set_xticklabels(xticks)
        
        _plt_open = pxpddict['Open'][_plot_ticker].iloc[_plot_range]
        _plt_high = pxpddict['High'][_plot_ticker].iloc[_plot_range]
        _plt_low = pxpddict['Low'][_plot_ticker].iloc[_plot_range]
        _plt_close = pxpddict['Close'][_plot_ticker].iloc[_plot_range]
        _px_colors = np.where(_plt_close>_plt_open, 'firebrick', 'darkslateblue')
        
        _tr = pxpddict['TurnoverRate'][_plot_ticker].iloc[_plot_range]
        
        try:
            axs[0].vlines(
                xs, 
                list(_plt_low.values), 
                list(_plt_high.values), 
                color=_px_colors,
                label=code_name_mapdict[_plot_ticker]['name'],
            )
        except:
            axs[0].vlines(
                    xs, 
                    list(_plt_low.values), 
                    list(_plt_high.values), 
                    color=_px_colors,
                )
        # plt.legend(loc=0, prop=fontP)
        axs[0].hlines(
            _plt_open, 
            np.array(xvs) - _ber_length, 
            xvs,
            color=_px_colors,
        )
        axs[0].hlines(
            _plt_close, 
            xvs, 
            np.array(xvs) + _ber_length,
            color=_px_colors,
        )
        
        try:
            axs[0].plot(xs, xpd_ma20[_plot_ticker].iloc[_plot_range].values, color='orange')
        #     axs.plot(xs, xpd_ma10[_plot_ticker].iloc[_plot_range].values, color='grey')
            axs[0].plot(xs, xpd_ma5[_plot_ticker].iloc[_plot_range].values, color='grey')
            
            axs[0].plot(xs, b_up[_plot_ticker].iloc[_plot_range].values, color='blue', alpha=0.5)
            axs[0].plot(xs, b_low[_plot_ticker].iloc[_plot_range].values, color='purple', alpha=0.5)
        except:
            pass

        _plt_ema12 = pxpddict['Close'][_plot_ticker].ewm(span=12, adjust=False).mean()
        _plt_ema26 = pxpddict['Close'][_plot_ticker].ewm(span=26, adjust=False).mean()
        _plt_ema60 = pxpddict['Close'][_plot_ticker].ewm(span=60, adjust=False).mean()
        _plt_ema130 = pxpddict['Close'][_plot_ticker].ewm(span=130, adjust=False).mean()
        _plt_ema260 = pxpddict['Close'][_plot_ticker].ewm(span=260, adjust=False).mean()
        
        axs[0].plot(xvs, _plt_ema12.iloc[_plot_range], color='red', alpha=0.2)
        axs[0].plot(xvs, _plt_ema26.iloc[_plot_range], color='blue', alpha=0.2)
        axs[0].plot(xvs, _plt_ema60.iloc[_plot_range], color='orange', alpha=0.2)
        axs[0].plot(xvs, _plt_ema130.iloc[_plot_range], color='green', alpha=0.2)
        axs[0].plot(xvs, _plt_ema260.iloc[_plot_range], color='purple', alpha=0.2)

        if alertedindo is not None:
            # st.table(alertedindo[_plot_ticker].iloc[-len(xs):])
            # aids = [i for i in range(len(xs)) if alertedindo[_plot_ticker].iloc[-len(xs):].iloc[i]]
            _xs = [x for x in xs if int(x.replace("-", "")) in alertedindo.index]
            aids = [i for i in range(len(_xs)) if alertedindo[_plot_ticker].loc[int(_xs[i].replace("-", ""))]]
            # def _get_alert_state(x):
            #     try:
            #         if alertedindo.loc[int(x.replace("-", ""))][_plot_ticker]:
            #             return xs.index(x)
            #     except:
            #         pass
            # aids = [_get_alert_state(x) for x in xs]
            # st.write(aids)
            axs[0].scatter([xs[i] for i in aids], [pxpddict['Close'][_plot_ticker].iloc[_plot_range].iloc[i] for i in aids], c='red', alpha=0.5, s=100)

        axs_0 = axs[0].twinx()
        axs_0.vlines(obs_step, 0, 1, linestyles='--', colors='black')
        axs_0.vlines(len(xs) - obs_step, 0, 1, linestyles='--', colors='black')

        # st.pyplot(fig, use_container_width=False)

        # if savepath is not None:
        #     try:
        #         fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}_{code_name_mapdict[_plot_ticker]['name']}.png", bbox_inches='tight')
        #     except:
        #         fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}.png", bbox_inches='tight')
        
        _plt_volume = pxpddict['TurnoverRate'][_plot_ticker].iloc[_plot_range]
        _plt_pctchg = pxpddict['Close'][_plot_ticker].pct_change().iloc[_plot_range]
        volcolors = np.where(_plt_pctchg>0, 'red', 'green')
        axs[1].bar(xvs, _plt_volume, color=volcolors)


        # fig, axs = plt.subplots(figsize=(16, 3))
        l1pd, l2pd, lhistpd = talib.MACD(xpd[_plot_ticker], fastperiod=12, slowperiod=26, signalperiod=9)
        axs[2].plot(xvs, l1pd.iloc[_plot_range], color='orange')
        axs[2].plot(xvs, l2pd.iloc[_plot_range], color='blue')
        axs[2].bar(xvs, lhistpd.iloc[_plot_range], color='blue')


        

        # try:
        #     axs[3].vlines(
        #         xs, 
        #         list(_plt_low.values), 
        #         list(_plt_high.values), 
        #         color=_px_colors,
        #         label=code_name_mapdict[_plot_ticker]['name'],
        #     )
        # except:
        #     axs[3].vlines(
        #             xs, 
        #             list(_plt_low.values), 
        #             list(_plt_high.values), 
        #             color=_px_colors,
        #         )
        # # plt.legend(loc=0, prop=fontP)
        # axs[3].hlines(
        #     _plt_open, 
        #     np.array(xvs) - _ber_length, 
        #     xvs,
        #     color=_px_colors,
        # )
        # axs[3].hlines(
        #     _plt_close, 
        #     xvs, 
        #     np.array(xvs) + _ber_length,
        #     color=_px_colors,
        # )

        import pandas as pd
        pd.DataFrame().ewm(span=12, adjust=False)

        # if savepath is not None:
        #     fig.savefig(f"{savepath}/{xxid}_{_plot_ticker}_macd.png", bbox_inches='tight')
        
        # if show:
            # plt.show()
        st.pyplot(fig, use_container_width=False)

        xxid+=1


from ibapi.client import *
from ibapi.wrapper import *
import datetime
import time
import threading
from ibapi.ticktype import TickTypeEnum

port = 7497


class TestApp(EClient, EWrapper):
    def __init__(self):
        EClient.__init__(self, self)

    def nextValidId(self, orderId: OrderId):
        self.orderId = orderId
    
    def nextId(self):
        self.orderId += 1
        return self.orderId
    
    def error(self, reqId, errorCode, errorString, advancedOrderReject=""):
        print(f"reqId: {reqId}, errorCode: {errorCode}, errorString: {errorString}, orderReject: {advancedOrderReject}")
    
    def headTimestamp(self, reqId, headTimeStamp):
        print(headTimeStamp)
        print(datetime.datetime.fromtimestamp(int(headTimeStamp)))
        self.cancelHeadTimeStamp(reqId)
        
    def tickPrice(self, reqId, tickType, price, attrib):
        print(f"reqId: {reqId}, tickType: {TickTypeEnum.toStr(tickType)}, price: {price}, attrib: {attrib}")

    def tickSize(self, reqId, tickType, size):
        print(f"reqId: {reqId}, tickType: {TickTypeEnum.toStr(tickType)}, size: {size}")



# app = TestApp()
# app.connect("127.0.0.1", port, 0)
# threading.Thread(target=app.run).start()
# time.sleep(1)

# mycontract = Contract()
# mycontract.symbol = "AAPL"
# mycontract.secType = "STK"
# mycontract.exchange = "SMART"
# mycontract.currency = "USD"

# # app.reqHeadTimeStamp(app.nextId(), mycontract, "TRADES", 1, 2)
# app.reqMarketDataType(3)
# app.reqMktData(app.nextId(), mycontract, "232", False, False, [])

# app.reqMktData()


import numpy as np 
import cv2 


# Capturing video through webcam 
webcam = cv2.VideoCapture(0) 

# Start a while loop 
while(1): 
	
	# Reading the video from the 
	# webcam in image frames 
	_, imageFrame = webcam.read() 

	# Convert the imageFrame in 
	# BGR(RGB color space) to 
	# HSV(hue-saturation-value) 
	# color space 
	hsvFrame = cv2.cvtColor(imageFrame, cv2.COLOR_BGR2HSV) 

	# Set range for red color and 
	# define mask 
	red_lower = np.array([136, 87, 111], np.uint8) 
	red_upper = np.array([180, 255, 255], np.uint8) 
	red_mask = cv2.inRange(hsvFrame, red_lower, red_upper) 

	# Set range for green color and 
	# define mask 
	green_lower = np.array([25, 52, 72], np.uint8) 
	green_upper = np.array([102, 255, 255], np.uint8) 
	green_mask = cv2.inRange(hsvFrame, green_lower, green_upper) 

	# Set range for blue color and 
	# define mask 
	blue_lower = np.array([94, 80, 2], np.uint8) 
	blue_upper = np.array([120, 255, 255], np.uint8) 
	blue_mask = cv2.inRange(hsvFrame, blue_lower, blue_upper) 
	
	# Morphological Transform, Dilation 
	# for each color and bitwise_and operator 
	# between imageFrame and mask determines 
	# to detect only that particular color 
	kernel = np.ones((5, 5), "uint8") 
	
	# For red color 
	red_mask = cv2.dilate(red_mask, kernel) 
	res_red = cv2.bitwise_and(imageFrame, imageFrame, 
							mask = red_mask) 
	
	# For green color 
	green_mask = cv2.dilate(green_mask, kernel) 
	res_green = cv2.bitwise_and(imageFrame, imageFrame, 
								mask = green_mask) 
	
	# For blue color 
	blue_mask = cv2.dilate(blue_mask, kernel) 
	res_blue = cv2.bitwise_and(imageFrame, imageFrame, 
							mask = blue_mask) 

	# Creating contour to track red color 
	contours, hierarchy = cv2.findContours(red_mask, 
										cv2.RETR_TREE, 
										cv2.CHAIN_APPROX_SIMPLE) 
	
	for pic, contour in enumerate(contours): 
		area = cv2.contourArea(contour) 
		if(area > 300): 
			x, y, w, h = cv2.boundingRect(contour) 
			imageFrame = cv2.rectangle(imageFrame, (x, y), 
									(x + w, y + h), 
									(0, 0, 255), 2) 
			
			cv2.putText(imageFrame, "Red Colour", (x, y), 
						cv2.FONT_HERSHEY_SIMPLEX, 1.0, 
						(0, 0, 255))	 

	# Creating contour to track green color 
	contours, hierarchy = cv2.findContours(green_mask, 
										cv2.RETR_TREE, 
										cv2.CHAIN_APPROX_SIMPLE) 
	
	for pic, contour in enumerate(contours): 
		area = cv2.contourArea(contour) 
		if(area > 300): 
			x, y, w, h = cv2.boundingRect(contour) 
			imageFrame = cv2.rectangle(imageFrame, (x, y), 
									(x + w, y + h), 
									(0, 255, 0), 2) 
			
			cv2.putText(imageFrame, "Green Colour", (x, y), 
						cv2.FONT_HERSHEY_SIMPLEX, 
						1.0, (0, 255, 0)) 

	# Creating contour to track blue color 
	contours, hierarchy = cv2.findContours(blue_mask, 
										cv2.RETR_TREE, 
										cv2.CHAIN_APPROX_SIMPLE) 
	for pic, contour in enumerate(contours): 
		area = cv2.contourArea(contour) 
		if(area > 300): 
			x, y, w, h = cv2.boundingRect(contour) 
			imageFrame = cv2.rectangle(imageFrame, (x, y), 
									(x + w, y + h), 
									(255, 0, 0), 2) 
			
			cv2.putText(imageFrame, "Blue Colour", (x, y), 
						cv2.FONT_HERSHEY_SIMPLEX, 
						1.0, (255, 0, 0)) 
			
	# Program Termination 
	cv2.imshow("Multiple Color Detection in Real-TIme", imageFrame) 
	if cv2.waitKey(10) & 0xFF == ord('q'): 
		webcam.release() 
		cv2.destroyAllWindows() 
		break


import signal
from contextlib import contextmanager
import logging
import logging.config

class TimeoutException(Exception): pass

@contextmanager
def exe_time_limit(secs):
    def signal_handler(signum, frame):
        raise TimeoutException("Time Out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(secs)
    try:
        yield
    finally:
        signal.alarm(0)

        
def free_memory(objlist):
    import gc
    if gc.collect() > 0:
        try:
            for xobj in objlist:
                del xobj
            _m = gc.collect()
            logging.info("Freed memory to {}".format(_m))
        except:
            pass


import os
import sys

# # icloud_froot = "/Users/{}/Library/Mobile Documents/com~apple~CloudDocs/".format(os.environ['USER'])
# icloud_froot = "/Users/{}/Documents/anaconda/anaconda3/lib/python3.11/site-packages/tsapi/_local/".format(os.environ['USER'])
# # sys.path.insert(0, icloud_froot + "/env/pkgs")

icloud_froot = "/Users/{}/Documents/pypi_projects/tsapix/_local/".format(os.environ['USER'])

env_db_root = icloud_froot + "env/data/"
cube_warehouse_root = icloud_froot + "env/data/cube/"
_log_froot = icloud_froot + "/env/log/"

env_raw_root = icloud_froot + "env/data/raw/"

code_univ_save_path = env_db_root + 'a_code_univ.yaml'

import requests
import json

import os
from typing import Dict, List
import pandas as pd
import numpy as np
import ruamel_yaml as yaml
import pyarrow.parquet as pq
import logging
import logging.config

from datetime import date

import numpy as np
import pandas as pd
import multitasking as _multitasking
import pandas_market_calendars as mcal

import pandas as _pd
import numpy as _np
import sys as _sys

def empty_df(index=None, cols=[], index_name=None):
    if index is None:
        index = []
    empty = _pd.DataFrame(index=index, data={c:_np.nan for c in cols})
    if index_name:
        empty.index.name = index_name
    return empty


class ProgressBar:
    def __init__(self, iterations, text='completed'):
        self.text = text
        self.iterations = iterations
        self.prog_bar = '[]'
        self.fill_char = '*'
        self.width = 50
        self.__update_amount(0)
        self.elapsed = 1

    def completed(self):
        if self.elapsed > self.iterations:
            self.elapsed = self.iterations
        self.update_iteration(1)
        print('\r' + str(self), end='', file=_sys.stderr)
        _sys.stderr.flush()
        print("", file=_sys.stderr)

    def animate(self, iteration=None):
        if iteration is None:
            self.elapsed += 1
            iteration = self.elapsed
        else:
            self.elapsed += iteration

        print('\r' + str(self), end='', file=_sys.stderr)
        _sys.stderr.flush()
        self.update_iteration()

    def update_iteration(self, val=None):
        val = val if val is not None else self.elapsed / float(self.iterations)
        self.__update_amount(val * 100.0)
        self.prog_bar += f"  {self.elapsed} of {self.iterations} {self.text}"

    def __update_amount(self, new_amount):
        percent_done = int(round((new_amount / 100.0) * 100.0))
        all_full = self.width - 2
        num_hashes = int(round((percent_done / 100.0) * all_full))
        self.prog_bar = '[' + self.fill_char * num_hashes + ' ' * (all_full - num_hashes) + ']'
        pct_place = (len(self.prog_bar) // 2) - len(str(percent_done))
        pct_string = f'{percent_done}%%'
        self.prog_bar = self.prog_bar[0:pct_place] + (pct_string + self.prog_bar[pct_place + len(pct_string):])

    def __str__(self):
        return str(self.prog_bar)


def save_as_yaml(yaml_path, content):
    with open(yaml_path, "w", encoding="utf-8") as f:
        yaml.dump(content, f, Dumper=yaml.RoundTripDumper)

def read_yaml_file(yaml_path):
    with open(yaml_path, "r") as f:
        yaml_dict = yaml.load(f.read(), Loader=yaml.Loader)
    return yaml_dict

def pq_file_to_pd(fpath):
    return pq.read_table(fpath).to_pandas()

def update_yaml(yaml_path, newpart, unique=False):
    try:
        histpart = read_yaml_file(yaml_path)
        if type(histpart) is List:
            newpart = histpart.extend(newpart)
            if unique:
                newpart = list(set(newpart))
        elif type(histpart) is Dict:
            newpart.update(histpart)
    except Exception as e:
        print(e)
        pass
    print("update_yaml  ", yaml_path)
    save_as_yaml(yaml_path, newpart)

def save_npcube(fpath, xcube):
    if ".npy" in fpath:
        np.save(fpath, xcube)
    elif ".npz" in fpath:
        np.savez_compressed(fpath, xcube)

def read_npcube(fpath):
    if ".npy" in fpath:
        return np.load(fpath)
    elif ".npz" in fpath:
        return np.load(fpath)['arr_0']

def nan_argsort(x, ascending=True):
    _mask = np.isnan(x.values)
    _argx = np.argsort(np.argsort(x.values[~_mask]))
    if not ascending:
        _argx = len(_argx) - _argx
    _x = x.values
    _x[~_mask] = _argx
    return _x

def initiate_logger(logfname, logfroot="", xlevel=logging.DEBUG):
    logging.basicConfig(filename=f"{logfroot}{logfname}.log", level=xlevel, 
                        format='%(asctime)s - %(levelname)s - %(message)s'
                        )
    return logging.getLogger(f"{logfroot}{logfname}.log")

def initiate_logger_v2(logfname, logfroot="", xlevel=logging.DEBUG, fmode='a'):
    logging.basicConfig(level=xlevel, 
                        format='%(asctime)s - %(levelname)s - %(message)s',
                        handlers=[
                            logging.FileHandler(filename=f"{logfroot}{logfname}.log", mode=fmode)
                        ]
                        )
    return logging.getLogger(f"{logfroot}{logfname}.log")

def save_cube_n_axes(xcube, xaxes, fname, save_root=None):
    save_npcube(f"{save_root}/{fname}.npz", xcube)
    save_as_yaml(f"{save_root}/{fname}.yaml", xaxes)
    
def load_cube_n_axes(pairname=None, froot=None, timeformat="%Y-%m-%d %H:%M"):
    xcube = read_npcube(f"{froot}/{pairname}.npz")
    xaxes = read_yaml_file(f"{froot}/{pairname}.yaml")
    xaxes[1] = pd.to_datetime(xaxes['1str'], format=timeformat)
    return {
        'cube': xcube,
        'axes': xaxes,
    }



from . import version
from . import eastmoney
from .base import _sv
# from .eastmoney import eastmoney_histpx_downloader, eastmoney_histpx_downloader_panellist
# from .eastmoney import eastmoney_histcf_downloader
# from .eastmoney import HistPX, HistCapitalFlow
# from .eastmoney import eastmoney_xbk_info_downloader, _retrive_perfinfo_at_specific_date, get_dtbk
# from .eastmoney import query_period_cf_allstock, query_period_cf_allstock_10d, query_period_cf_allstock_3d, query_period_cf_allstock_5d
# from .eastmoney import query_period_cf_allstock_rtn

__version__ = version
__author__ = 'TS'


__all__ = []

__all__.extend(eastmoney.__all__)

# __all__ = [
#     # 'eastmoney_histpx_downloader',
#     # 'eastmoney_histpx_downloader_panellist',
#     # 'eastmoney_histcf_downloader'
#     'HistPX',
#     'HistCapitalFlow',
#     '_sv',
#     'eastmoney_xbk_info_downloader',
#     '_retrive_perfinfo_at_specific_date',
#     'get_dtbk',
#     'query_period_cf_allstock',
#     'query_period_cf_allstock_10d',
#     'query_period_cf_allstock_3d',
#     'query_period_cf_allstock_5d',

#     'query_period_cf_allstock_rtn',

# ]




import os
import time
from tsapix.utils import icloud_froot, env_db_root, cube_warehouse_root, _log_froot, env_raw_root
import tsapix
from tsapix.utils import *

#1.##################### initiate folders to store data ######################
print("###################### Initiate folders to store data...")
for fld in [icloud_froot, env_db_root, cube_warehouse_root, _log_froot, env_raw_root]:
    os.makedirs(fld, exist_ok=True)
    print(fld)

time.sleep(5)


#2.##################### Initiate China tickers ######################
print("###################### Initiate China tickers ...")
from tsapix.spider.ticker_cn_update import update_cn_ticker_universe
import tsapix.spider.eastmoney as cnsource

try:
    update_cn_ticker_universe()

    cnsource.eastmoney_query_ST_names(savepath=env_db_root)

except:
    print("Failed!")

#2.1##################### China tickers meta data ######################
print("###################### Meata data: Industry")
cnsource.eastmoney_industry_info_downloader()
print("###################### Meata data: Concept")
cnsource.eastmoney_concept_info_downloader()
print("###################### Meata data: Geo")
cnsource.eastmoney_geo_info_downloader()

print("###################### Generating Meata data Dict...")
a_univ_list = read_yaml_file(code_univ_save_path)
cnsource.get_stock_meta_data(a_univ_list, dbroot=env_raw_root, save_dbroot=env_db_root)


# #3.##################### Initiate China Data Base ######################
# print("###################### Initiate China Data Base ...")
# from tsapix.tools.taskbase import exe_time_limit, free_memory
# import logging
# import gc

# import tsapix.spider.eastmoney as cnsource


# tuiv = read_yaml_file(code_univ_save_path)
# tuiv = tuiv[:5]
# print("Number of China tickers:", len(tuiv))
# single_task_timelmt = 60 * 60 * 10

# for a, b in [
#     ('d', 'bfq'), # meaning [daily, unadjusted]
#     # ('30min', 'bfq'),
#     # ('60min', 'bfq'),
# ]:
#     print("Ingesting data @", a, b)
#     try:
#         with exe_time_limit(single_task_timelmt):
#             xobj = cnsource.HistPX(a, b, tickers=tuiv, start="2011-12-31", end=date.today().strftime("%Y-%m-%d"), lmt='1000000')
#             xobj.download(chunk_size=200)

#             xobj.get_cube_and_save(pairname=f'histpx_{a}_{b}', froot=cube_warehouse_root)
#     except Exception as e:
#         logging.error(e)

#     free_memory([xobj,])
#     time.sleep(40)

# import beepy
# beepy.beep(6)


from .utils import *

try:
    a_code_univlist = read_yaml_file(code_univ_save_path)
except:
    a_code_univlist = []
sz_identifier_0 = ['0', '3'] # listed in Shen Zhen if ticker start with 0 or 3


histpx_cols = ['Open', 'Close', 'High', 'Low', 'TradeShares', 'TradeAmt', 'ZF', 'AbsChg', 'PctChg', 'TurnoverRate']
histcf_cols = ['zljlr_je','xdjlr_je','zdjlr_je','ddjlr_je','cddjlr_je','zljlr_jzb','xdjlr_jzb','zdjlr_jzb','ddjlr_jzb','cddjlr_jzb','Close','PctChg','nan1','nan2']

china_tcalendar = mcal.get_calendar('SSE')

def _continuous_auction_period(x):
    return np.any([
            np.all([x.strftime("%H:%M:%S")<="15:00:00", x.strftime("%H:%M:%S")>="13:00:00"], axis=0),
            np.all([x.strftime("%H:%M:%S")<="11:30:00", x.strftime("%H:%M:%S")>="09:30:00"], axis=0),
        ], axis=0)

def _call_auction_period(x):
    return np.all([x.strftime("%H:%M:%S")>="09:15:00", x.strftime("%H:%M:%S")<="09:25:00"], axis=0)

def _sv(xobj, a0=0,): # slicing view
    return pd.DataFrame(xobj.xcube[a0, :, :], index=xobj.axsesinfo[1], columns=xobj.axsesinfo[2])

_eastmoney_filed_meaningdict = {
    
}


def get_stock_meta_dict():
    """ Dict record stock meta info:

    Element Structure
    '000001': {
        'name': 'xxx',
        'industry': [],
        'geo': [],
        'concept: [],
    }
    
    """
    return read_yaml_file(env_db_root + "stock_meta_dict.yaml")

def get_stock_codechinesemap():
    """ Dict record stock code&name info:

    Element Structure
    '000001': 'xxx'
    
    """
    return read_yaml_file(env_db_root + "stock_codechinesemap.yaml")

def get_bk_namemap():
    """ Dict record bankuai(sector) code&name info:

    Element Structure
    'BKXXX': 'xxx'
    
    """
    return read_yaml_file(env_db_root + "bk_namemap.yaml")

def load_maps():
    xdict = {
        'stock_meta': get_stock_meta_dict(),
        'bkname': get_bk_namemap(),
    }
    print("Maps Available Keys: ", list(xdict.keys()))
    return xdict

def load_cube_n_axes(pairname=None, froot=None, timeformat="%Y-%m-%d %H:%M"):
    xcube = read_npcube(f"{froot}/{pairname}.npz")
    xaxes = read_yaml_file(f"{froot}/{pairname}.yaml")
    xaxes[1] = pd.to_datetime(xaxes['1str'], format=timeformat)
    return {
        'cube': xcube,
        'axes': xaxes,
    }

def get_dtbk(start, end, xfreq, 
        _freq_mapdict={
            'd': "D",
            'w': "W",
            'm': "M",
            '5min': "5min",
            '15min': "15min",
            '30min': "30min",
            '60min': "60min",
    }, tcalendar=china_tcalendar, ):
    trading_dates = [x.strftime("%Y-%m-%d") for x in tcalendar.valid_days(start_date=start, end_date=end)]
    _valid_list = pd.bdate_range(start, end, freq=_freq_mapdict[xfreq])

    if 'min' in xfreq or 's' in xfreq:
        # _in_trading_times
        _valid_list = [x for x in _valid_list if np.any([
            np.all([x.strftime("%H:%M:%S")<="15:00:00", x.strftime("%H:%M:%S")>="13:00:00"], axis=0),
            np.all([x.strftime("%H:%M:%S")<="11:30:00", x.strftime("%H:%M:%S")>="09:30:00"], axis=0),
        ], axis=0)]
    # _in_trading_dates
    _valid_list = [x for x in _valid_list if x.strftime("%Y-%m-%d") in trading_dates]
    
    if xfreq in ['w', 'm']:
        def _xdate(x):
            try:
                return x.index[-1].strftime("%Y-%m-%d")
            except:
                return np.nan
        _valid_list = list(pd.DataFrame({'dts': pd.to_datetime(trading_dates), 'dts0': pd.to_datetime(trading_dates)}).set_index('dts').resample(_freq_mapdict[xfreq]).apply(lambda x: _xdate(x)).dropna().values[:, 0])

    return pd.DataFrame({0:_valid_list}).set_index(0).sort_index()



import requests
import json
import time
import  re
# import pymssql
# import pymysql
import  datetime
import random
from requests.adapters import HTTPAdapter
import pandas as pd
import numpy as np
from functools import reduce
import matplotlib.pyplot as plt

datapath = "./"

import logging
LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
logging.basicConfig(filename=f'{datapath}sina_hourlyupdates.log', level=logging.DEBUG, format=LOG_FORMAT)


class Colour:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'


def screening_keywords(keyword=''):
    page = 0
    while True:
        try:
            page+=1
            referer_url = "http://finance.sina.com.cn/7x24/?tag=0"
            cookie = "UOR=www.baidu.com,tech.sina.com.cn,; SINAGLOBAL=114.84.181.236_1579684610.152568; UM_distinctid=16fcc8a8b704c8-0a1d2def9ca4c6-33365a06-15f900-16fcc8a8b718f1; lxlrttp=1578733570; gr_user_id=2736e487-ee25-4d52-a1eb-c232ac3d58d6; grwng_uid=d762fe92-912b-4ea8-9a24-127a43143ebf; __gads=ID=d79f786106eb99a1:T=1582016329:S=ALNI_MZoErH_0nNZiM3D4E36pqMrbHHOZA; Apache=114.84.181.236_1582267433.457262; ULV=1582626620968:6:4:1:114.84.181.236_1582267433.457262:1582164462661; ZHIBO-SINA-COM-CN=; SUB=_2AkMpBPEzf8NxqwJRmfoWz2_ga4R2zQzEieKfWADoJRMyHRl-yD92qm05tRB6AoTf3EaJ7Bg2UU4l1CDZXUBCzEuJv3mP; SUBP=0033WrSXqPxfM72-Ws9jqgMF55529P9D9WhqhhGsPWdPjar0R99pFT8s"
            headers = {
                "Accept": "*/*",
                "Accept-Encoding": "gzip, deflate",
                "Accept-Language": "zh-CN,zh;q=0.9",
                "Connection": "keep-alive",
                "Cookie": cookie,
                "Host": "zhibo.sina.com.cn",
                "Referer": referer_url,
                "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36"
            }
            base_url = "http://zhibo.sina.com.cn/api/zhibo/feed?callback=jQuery0&page=%s"%page+"&page_size=20&zhibo_id=152&tag_id=0&dire=f&dpc=1&pagesize=20&_=0%20Request%20Method:GET%27"
            data = get_json_data(base_url,headers)
            
            time.sleep(10)
            print("+", end=' ')
        except Exception as e:
            print('error', e)
            continue
            
def get_json_data(base_url,headers):
    s = requests.Session()
    s.mount('http://', HTTPAdapter(max_retries=3))
    s.mount('https://', HTTPAdapter(max_retries=3))
    print(time.strftime('%Y-%m-%d %H:%M:%S'))
 
    try:
        response = requests.get(base_url, timeout=5, headers=headers)
        html = response.text
        # print(html)
        html_cl = html[12:-14]
        false = False
        true = True
        null = None
        html_json = eval(html_cl)
        json_str = json.dumps(html_json)
        results = json.loads(json_str)
        data = results['result']['data']['feed']['list']
    except Exception as e:
        print('get_json_str未收录错误类型，请检查网络通断,错误位置：',e)
        time.sleep(5)
        get_json_data(base_url, headers)
    else:
        return data
    
def get_info_from_json(jsonlist):
    _resdict = {
        "id": [],
        "txt": [],
        "time": [],
    }
    _tradable = {
        "market": [],
        "symbol": [],
        "key": [],
        "id": []
    }
    for d in jsonlist:
        _resdict["id"].append(d["id"])
        _resdict["txt"].append(d["rich_text"])
        _resdict["time"].append(d["create_time"])
        try:
            _ext = json.loads(d['ext'])
            if len(_ext['stocks'])>0:
                for s in _ext['stocks']:
                    _tradable['market'].append(s['market'])
                    _tradable['symbol'].append(s['symbol'])
                    _tradable['key'].append(s['key'])
                    _tradable['id'].append(d["id"])
        except Exception as e:
            logging.error(e)
    return _resdict, _tradable


class ScreeningKeyword:
    
    def __init__(self):
        self.news_bag = []
    
    def screening_keywords(self, keyword='', record_target=100, start_page=0, time_limit='2010-03-19 18:38:45'):
        page = start_page
        while True:
            try:
                page+=1
                referer_url = "http://finance.sina.com.cn/7x24/?tag=0"
                cookie = "UOR=www.baidu.com,tech.sina.com.cn,; SINAGLOBAL=114.84.181.236_1579684610.152568; UM_distinctid=16fcc8a8b704c8-0a1d2def9ca4c6-33365a06-15f900-16fcc8a8b718f1; lxlrttp=1578733570; gr_user_id=2736e487-ee25-4d52-a1eb-c232ac3d58d6; grwng_uid=d762fe92-912b-4ea8-9a24-127a43143ebf; __gads=ID=d79f786106eb99a1:T=1582016329:S=ALNI_MZoErH_0nNZiM3D4E36pqMrbHHOZA; Apache=114.84.181.236_1582267433.457262; ULV=1582626620968:6:4:1:114.84.181.236_1582267433.457262:1582164462661; ZHIBO-SINA-COM-CN=; SUB=_2AkMpBPEzf8NxqwJRmfoWz2_ga4R2zQzEieKfWADoJRMyHRl-yD92qm05tRB6AoTf3EaJ7Bg2UU4l1CDZXUBCzEuJv3mP; SUBP=0033WrSXqPxfM72-Ws9jqgMF55529P9D9WhqhhGsPWdPjar0R99pFT8s"
                headers = {
                    "Accept": "*/*",
                    "Accept-Encoding": "gzip, deflate",
                    "Accept-Language": "zh-CN,zh;q=0.9",
                    "Connection": "keep-alive",
                    "Cookie": cookie,
                    "Host": "zhibo.sina.com.cn",
                    "Referer": referer_url,
                    "User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.117 Safari/537.36"
                }
                base_url = "http://zhibo.sina.com.cn/api/zhibo/feed?callback=jQuery0&page=%s"%page+"&page_size=20&zhibo_id=152&tag_id=0&dire=f&dpc=1&pagesize=20&_=0%20Request%20Method:GET%27"
                self.data = get_json_data(base_url,headers)
                self.related_data = [e for e in self.data if keyword in e['rich_text']]
                self.news_bag.extend(self.related_data)

                time.sleep(10)
                print("@", page, self.data[-1]['update_time'], end=' ')
                print("News Bag Length: ", len(self.news_bag), "/", record_target)
                
                if len(self.news_bag) >= record_target:
                    print('Reach target limit... Exit')
                    break
                    
                if self.data[-1]['update_time'] < time_limit:
                    break
                    
            except Exception as e:
                print('error', e)
                continue
                
            except KeyboardInterrupt:
                break
                
                

def normal_task_process(keyword='加沙', record_target=100, start_page=0, time_limit='2010-03-19 18:38:45'):
    xobj = ScreeningKeyword()
    xobj.screening_keywords(keyword=keyword, record_target=record_target, start_page=start_page, time_limit=time_limit)

    try:
        _resdict, _tradable = get_info_from_json(xobj.news_bag)
        # pd.DataFrame(_resdict).sort_values('id')
        return _resdict
    except Exception as e:
        print("Error... Returning xobj, use xobj.news_bag to check results.")
        return xobj